##from fipy.tools.profiler.profiler import Profiler
##from fipy.tools.profiler.profiler import calibrate_profiler

"""

This input file again solves a 1D diffusion problem as in
`./examples/diffusion/steadyState/mesh1D/input.py`. The difference
being that it uses a triangular mesh loaded in using the
:class:`~fipy.meshes.gmshMesh.Gmsh2D` object.

The result is again tested in the same way:

>>> from fipy import Gmsh2D, CellVariable, DiffusionTerm, Viewer
>>> from fipy.tools import numerix

>>> valueLeft = 0.
>>> valueRight = 1.

>>> Lx = 20
>>> mesh = Gmsh2D('''
...     cellSize = 0.5;
...     Point(2) = {0, 0, 0, cellSize};
...     Point(3) = {%(Lx)g, 0, 0, cellSize};
...     Point(4) = {%(Lx)g, %(Lx)g, 0, cellSize};
...     Point(5) = {0, %(Lx)g, 0, cellSize};
...
...     Line(6) = {2, 3};
...     Line(7) = {3, 4};
...     Line(8) = {4, 5};
...     Line(9) = {5, 2};
...
...     Line Loop(10) = {6, 7, 8, 9};
...
...     Plane Surface(11) = {10};
...     ''' % locals()) # doctest: +GMSH

>>> var = CellVariable(name = "solution variable",
...               mesh = mesh,
...               value = valueLeft) # doctest: +GMSH

>>> var.constrain(valueLeft, mesh.facesLeft) # doctest: +GMSH
>>> var.constrain(valueRight, mesh.facesRight) # doctest: +GMSH

>>> DiffusionTerm().solve(var) # doctest: +GMSH

>>> from fipy import input
>>> x = mesh.cellCenters[0] # doctest: +GMSH
>>> analyticalArray = valueLeft + (valueRight - valueLeft) * x / Lx # doctest: +GMSH
>>> print(var.allclose(analyticalArray, atol=0.025)) # doctest: +GMSH
True

>>> errorVar = abs(var - analyticalArray) # doctest: +GMSH
>>> errorVar.name = "absolute error" # doctest: +GMSH

>>> NonOrthoVar = CellVariable(name="non-orthogonality",
...                            mesh=mesh,
...                            value=mesh._nonOrthogonality) # doctest: +GMSH
>>> print(max(NonOrthoVar) < 0.51) # doctest: +GMSH
True

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=var)
...     viewer.plot()
...
...     errorViewer = Viewer(vars=errorVar)
...     errorViewer.plot()
...
...     NOViewer = Viewer(vars=NonOrthoVar)
...     NOViewer.plot()
...
...     input("finished")
"""


__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



from fipy.tools import numerix

__all__ = ["OffsetSparseMatrix"]

def OffsetSparseMatrix(SparseMatrix,
                       numberOfVariables, numberOfEquations,
                       equationIndex, varIndex):
    """
    Used in binary terms. `equationIndex` and `varIndex` need to be set statically before instantiation.
    """

    class OffsetSparseMatrixClass(SparseMatrix):
        def __init__(self, mesh, nonZerosPerRow=1, exactNonZeros=False,
                     numberOfVariables=numberOfVariables,
                     numberOfEquations=numberOfEquations,
                     equationIndex=equationIndex,
                     varIndex=varIndex):
            self.equationIndex = equationIndex
            self.varIndex = varIndex
            if hasattr(nonZerosPerRow, "__iter__"):
                # nonZerosPerRow is an iterable for each row.
                # need to pad rows for other equations with zeros.
                # can't compare to collections.abc.Iterable because PySparse.
                before = self.equationIndex
                after = numberOfEquations - self.equationIndex - 1
                N = len(nonZerosPerRow)
                nonZerosPerRow = numerix.concatenate([[0] * N * before,
                                                      nonZerosPerRow,
                                                      [0] * N * after])
            else:
                nonZerosPerRow //= numberOfEquations
            SparseMatrix.__init__(self,
                                  mesh=mesh,
                                  nonZerosPerRow=nonZerosPerRow,
                                  exactNonZeros=exactNonZeros,
                                  numberOfVariables=numberOfVariables,
                                  numberOfEquations=numberOfEquations)

        def put(self, vector, id1, id2):
            N = self.mesh.numberOfCells
            SparseMatrix.put(self,
                             vector=vector,
                             id1=id1 + N * self.equationIndex,
                             id2=id2 + N * self.varIndex)

        def addAt(self, vector, id1, id2):
            N = self.mesh.numberOfCells
            SparseMatrix.addAt(self,
                               vector=vector,
                               id1=id1 + N * self.equationIndex,
                               id2=id2 + N * self.varIndex)

        def addAtDiagonal(self, vector):
            if type(vector) in [type(1), type(1.)]:
                tmp = numerix.zeros((self.mesh.numberOfCells,), 'd')
                tmp[:] = vector
                SparseMatrix.addAtDiagonal(self, tmp)
            else:
                SparseMatrix.addAtDiagonal(self, vector)

    return OffsetSparseMatrixClass



from PyTrilinos import AztecOO

from ..convergence import (Convergence, Divergence,
                           BreakdownDivergence,
                           IllConditionedDivergence,
                           IterationDivergence,
                           LossOfAccuracyConvergence)

class AZ_NormalConvergence(Convergence):
    """
    """
    status_code = AztecOO.AZ_normal
    status_name = "AZ_normal"
    suite = "trilinos"

# Is this a convergence or divergence?
# Does it mean the same thing as OutOfRangeDivergence?
class AZ_LossOfAccuracyConvergence(LossOfAccuracyConvergence):
    """Numerical loss of precision occurred.
    """
    status_code = AztecOO.AZ_loss
    status_name = "AZ_loss"
    suite = "trilinos"

class AZ_ParameterDivergence(Divergence):
    """
    """
    status_code = AztecOO.AZ_param
    status_name = "AZ_param"
    suite = "trilinos"

class AZ_BreakdownDivergence(BreakdownDivergence):
    """
    """
    status_code = AztecOO.AZ_breakdown
    status_name = "AZ_breakdown"
    suite = "trilinos"

class AZ_IllConditionedDivergence(IllConditionedDivergence):
    """
    """
    status_code = AztecOO.AZ_ill_cond
    status_name = "AZ_ill_cond"
    suite = "trilinos"

class AZ_IterationDivergence(IterationDivergence):
    """
    """
    status_code = AztecOO.AZ_maxits
    status_name = "AZ_maxits"
    suite = "trilinos"



r"""
This example combines a phase field problem, as given in
:mod:`examples.elphf.phase`, with a binary diffusion problem, such as
described in the ternary example :mod:`examples.elphf.diffusion.mesh1D`, on a
1D mesh

>>> from fipy import CellVariable, FaceVariable, Grid1D, TransientTerm, DiffusionTerm, ImplicitSourceTerm, PowerLawConvectionTerm, Viewer
>>> from fipy.tools import numerix

>>> nx = 400
>>> dx = 0.01
>>> L = nx * dx
>>> mesh = Grid1D(dx = dx, nx = nx)

We create the phase field

>>> phase = CellVariable(mesh=mesh, name='xi', value=1., hasOld=1)
>>> phase.mobility = 1.
>>> phase.gradientEnergy = 0.025

>>> def pPrime(xi):
...     return 30. * (xi * (1 - xi))**2

>>> def gPrime(xi):
...     return 4 * xi * (1 - xi) * (0.5 - xi)

and a dummy electrostatic potential field

>>> potential = CellVariable(mesh = mesh, name = 'phi', value = 0.)
>>> permittivityPrime = 0.

We start with a binary substitutional system

>>> class ComponentVariable(CellVariable):
...     def __init__(self, mesh, value = 0., name = '',
...                  standardPotential = 0., barrier = 0.,
...                  diffusivity = None, valence = 0, equation = None,
...                  hasOld = 1):
...         self.standardPotential = standardPotential
...         self.barrier = barrier
...         self.diffusivity = diffusivity
...         self.valence = valence
...         self.equation = equation
...         CellVariable.__init__(self, mesh = mesh, value = value,
...                               name = name, hasOld = hasOld)
... 
...     def copy(self):
...         return self.__class__(mesh = self.mesh,
...                               value = self.value,
...                               name = self.name,
...                               standardPotential =
...                                   self.standardPotential,
...                               barrier = self.barrier,
...                               diffusivity = self.diffusivity,
...                               valence = self.valence,
...                               equation = self.equation,
...                               hasOld = 0)

consisting of the solvent

>>> solvent = ComponentVariable(mesh = mesh, name = 'Cn', value = 1.)

and the solute

>>> substitutionals = [
...     ComponentVariable(mesh = mesh, name = 'C1',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.3/.7) - numerix.log(.7/.3))]
>>> interstitials = []

>>> for component in substitutionals:
...     solvent -= component

The thermodynamic parameters are chosen to give a solid phase rich
in the solute and a liquid phase rich in the solvent.

.. warning: Addition and subtraction cause ``solvent`` to lose some crucial
   information so we only append it after the fact.

>>> solvent.standardPotential = numerix.log(.7/.3)
>>> solvent.barrier = 1.

We create the phase equation as in :mod:`examples.elphf.phase`
and create the diffusion equations for the different species as in
:mod:`examples.elphf.diffusion.mesh1D`.  The initial residual of the
diffusion equations is much larger than the norm of the right-hand-side
vector, so we use `"initial"` tolerance scaling for those equations

>>> def makeEquations(phase, substitutionals, interstitials):
...     phase.equation = TransientTerm(coeff = 1/phase.mobility) \
...         == DiffusionTerm(coeff = phase.gradientEnergy) \
...         - (permittivityPrime / 2.) \
...             * potential.grad.dot(potential.grad)
...     enthalpy = solvent.standardPotential
...     barrier = solvent.barrier
...     for component in substitutionals + interstitials:
...         enthalpy += component * component.standardPotential
...         barrier += component * component.barrier
... 
...     mXi = -(30 * phase * (1 - phase) * enthalpy
...             +  4 * (0.5 - phase) * barrier)
...     dmXidXi = (-60 * (0.5 - phase) * enthalpy + 4 * barrier)
...     S1 = dmXidXi * phase * (1 - phase) + mXi * (1 - 2 * phase)
...     S0 = mXi * phase * (1 - phase) - phase * S1
... 
...     phase.equation -= S0 + ImplicitSourceTerm(coeff = S1)
... 
...     for Cj in substitutionals:
...         CkSum = ComponentVariable(mesh = mesh, value = 0.)
...         CkFaceSum = FaceVariable(mesh = mesh, value = 0.)
...         for Ck in [Ck for Ck in substitutionals if Ck is not Cj]:
...             CkSum += Ck
...             CkFaceSum += Ck.harmonicFaceValue
... 
...         counterDiffusion = CkSum.faceGrad
...         phaseTransformation = (pPrime(phase.harmonicFaceValue) \
...                 * Cj.standardPotential
...                 + gPrime(phase.harmonicFaceValue) \
...                     * Cj.barrier) * phase.faceGrad
...         electromigration = Cj.valence * potential.faceGrad
...         convectionCoeff = counterDiffusion + \
...             solvent.harmonicFaceValue \
...                 * (phaseTransformation + electromigration)
...         convectionCoeff *= \
...             (Cj.diffusivity / (1. - CkFaceSum))
... 
...         Cj.equation = (TransientTerm()
...                        == DiffusionTerm(coeff=Cj.diffusivity)
...                        + PowerLawConvectionTerm(coeff=convectionCoeff))
... 
...     for Cj in interstitials:
...         phaseTransformation = (pPrime(phase.harmonicFaceValue) \
...             * Cj.standardPotential \
...             + gPrime(phase.harmonicFaceValue) \
...                 * Cj.barrier) * phase.faceGrad
...         electromigration = Cj.valence * potential.faceGrad
...         convectionCoeff = Cj.diffusivity \
...             * (1 + Cj.harmonicFaceValue) \
...             * (phaseTransformation + electromigration)
... 
...         Cj.equation = (TransientTerm()
...                        == DiffusionTerm(coeff=Cj.diffusivity)
...                        + PowerLawConvectionTerm(coeff=convectionCoeff))
...
...     for Cj in substitutionals + interstitials:
...         Cj.solver = Cj.equation.getDefaultSolver(criterion="initial",
...                                                  tolerance=1e-7)

>>> makeEquations(phase, substitutionals, interstitials)

We start with a sharp phase boundary

.. math::

   \xi =
   \begin{cases}
       1& \text{for $x \le L/2$,} \\
       0& \text{for $x > L/2$,}
   \end{cases}

or

>>> x = mesh.cellCenters[0]
>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L / 2)

and with a uniform concentration field :math:`C_1 = 0.5` or

>>> substitutionals[0].setValue(0.5)

If running interactively, we create viewers to display the results

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=([phase, solvent]
...                           + substitutionals + interstitials),
...                     datamin=0, datamax=1)
...     viewer.plot()

This problem does not have an analytical solution, so after
iterating to equilibrium

>>> dt = 10000
>>> for i in range(5):
...     for field in [phase] + substitutionals + interstitials:
...         field.updateOld()
...     phase.equation.solve(var = phase, dt = dt)
...     for field in substitutionals + interstitials:
...         field.equation.solve(var=field,
...                              dt=dt,
...                              solver=field.solver)
...     if __name__ == '__main__':
...         viewer.plot()

.. image:: /figures/examples/elphf/phaseDiffusion/binary.*
   :scale: 50
   :align: center
   :alt: phase and two concentration fields in equilibrium

we confirm that the far-field phases have remained separated

>>> numerix.allclose(phase(((0., L),)), (1.0, 0.0), rtol = 1e-5, atol = 1e-5)
1

and that the solute concentration field has appropriately segregated into
solute-rich and solute-poor phases.

>>> print(numerix.allclose(substitutionals[0](((0., L),)), (0.7, 0.3), rtol = 2e-3, atol = 2e-3))
1

The same system of equations can model a quaternary substitutional system as
easily as a binary.  Because it depends on the number of substitutional solute
species in question, we recreate the solvent

>>> solvent = ComponentVariable(mesh = mesh, name = 'Cn', value = 1.)

and make three new solute species

>>> substitutionals = [
...     ComponentVariable(mesh = mesh, name = 'C1',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.3/.4) - numerix.log(.1/.2)),
...     ComponentVariable(mesh = mesh, name = 'C2',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.4/.3) - numerix.log(.1/.2)),
...     ComponentVariable(mesh = mesh, name = 'C3',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.2/.1) - numerix.log(.1/.2))]

>>> for component in substitutionals:
...     solvent -= component
>>> solvent.standardPotential = numerix.log(.1/.2)
>>> solvent.barrier = 1.

These thermodynamic parameters are chosen to give a solid phase rich in the
solvent and the first substitutional component and a liquid phase rich in
the remaining two substitutional species.

Again, if we're running interactively, we create a viewer

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=([phase, solvent]
...                           + substitutionals + interstitials),
...                     datamin=0, datamax=1)
...     viewer.plot()

We reinitialize the sharp phase boundary

>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L / 2)

and the uniform concentration fields, with the substitutional concentrations
:math:`C_1 = C_2 = 0.35` and :math:`C_3 = 0.15`.

>>> substitutionals[0].setValue(0.35)
>>> substitutionals[1].setValue(0.35)
>>> substitutionals[2].setValue(0.15)

We make new equations

>>> makeEquations(phase, substitutionals, interstitials)

and again iterate to equilibrium

>>> dt = 10000
>>> for i in range(5):
...     for field in [phase] + substitutionals + interstitials:
...         field.updateOld()
...     phase.equation.solve(var = phase, dt = dt)
...     for field in substitutionals + interstitials:
...         field.equation.solve(var=field,
...                              dt=dt,
...                              solver=field.solver)
...     if __name__ == '__main__':
...         viewer.plot()

.. image:: /figures/examples/elphf/phaseDiffusion/quaternary.*
   :scale: 50
   :align: center
   :alt: phase and four concentration fields in equilibrium

We confirm that the far-field phases have remained separated

>>> numerix.allclose(phase(((0., L),)), (1.0, 0.0), rtol = 1e-5, atol = 1e-5)
1

and that the concentration fields have appropriately segregated into
their respective phases

>>> numerix.allclose(substitutionals[0](((0., L),)), (0.4, 0.3), rtol = 3e-3, atol = 3e-3)
1
>>> numerix.allclose(substitutionals[1](((0., L),)), (0.3, 0.4), rtol = 3e-3, atol = 3e-3)
1
>>> numerix.allclose(substitutionals[2](((0., L),)), (0.1, 0.2), rtol = 3e-3, atol = 3e-3)
1

Finally, we can represent a system that contains both substitutional and
interstitial species. We recreate the solvent

>>> solvent = ComponentVariable(mesh = mesh, name = 'Cn', value = 1.)

and two new solute species

>>> substitutionals = [
...     ComponentVariable(mesh = mesh, name = 'C2',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.4/.3) - numerix.log(.4/.6)),
...     ComponentVariable(mesh = mesh, name = 'C3',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.2/.1) - numerix.log(.4/.6))]

and one interstitial

>>> interstitials = [
...     ComponentVariable(mesh = mesh, name = 'C1',
...                       diffusivity = 1., barrier = 0.,
...                       standardPotential = numerix.log(.3/.4) - numerix.log(1.3/1.4))]


>>> for component in substitutionals:
...     solvent -= component
>>> solvent.standardPotential = numerix.log(.4/.6) - numerix.log(1.3/1.4)
>>> solvent.barrier = 1.

These thermodynamic parameters are chosen to give a solid phase rich in interstitials
and the solvent and a liquid phase rich in the two substitutional species.

Once again, if we're running interactively, we create a viewer

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=([phase, solvent]
...                           + substitutionals + interstitials),
...                     datamin=0, datamax=1)
...     viewer.plot()

We reinitialize the sharp phase boundary

>>> phase.setValue(1.)
>>> phase.setValue(0., where=x > L / 2)

and the uniform concentration fields, with the interstitial concentration
:math:`C_1 = 0.35`

>>> interstitials[0].setValue(0.35)

and the substitutional concentrations :math:`C_2 = 0.35` and :math:`C_3 = 0.15`.

>>> substitutionals[0].setValue(0.35)
>>> substitutionals[1].setValue(0.15)

We make new equations

>>> makeEquations(phase, substitutionals, interstitials)

and again iterate to equilibrium

>>> dt = 10000
>>> for i in range(5):
...     for field in [phase] + substitutionals + interstitials:
...         field.updateOld()
...     phase.equation.solve(var = phase, dt = dt)
...     for field in substitutionals + interstitials:
...         field.equation.solve(var=field,
...                              dt=dt,
...                              solver=field.solver)
...     if __name__ == '__main__':
...         viewer.plot()

.. image:: /figures/examples/elphf/phaseDiffusion/ternaryAndElectrons.*
   :scale: 50
   :align: center
   :alt: phase and four concentration fields (one like electrons) in equilibrium

We once more confirm that the far-field phases have remained separated

>>> numerix.allclose(phase(((0., L),)), (1.0, 0.0), rtol = 1e-5, atol = 1e-5)
1

and that the concentration fields have appropriately segregated into
their respective phases

>>> numerix.allclose(interstitials[0](((0., L),)), (0.4, 0.3), rtol = 3e-3, atol = 3e-3)
1
>>> numerix.allclose(substitutionals[0](((0., L),)), (0.3, 0.4), rtol = 3e-3, atol = 3e-3)
1
>>> numerix.allclose(substitutionals[1](((0., L),)), (0.1, 0.2), rtol = 3e-3, atol = 3e-3)
1
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == "__main__":
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())
    input("finished")

## if __name__ == '__main__':
##     ## from fipy.tools.profiler.profiler import Profiler
##     ## from fipy.tools.profiler.profiler import calibrate_profiler
##
##     # fudge = calibrate_profiler(10000)
##     # profile = Profiler('profile', fudge=fudge)
##
##     import fipy.tests.doctestPlus
##     exec(fipy.tests.doctestPlus.getScript())
##
##     # profile.stop()
##
##     raw_input("finished")
##



"""
It is straightforward to extend a phase field model to include binary alloys.
As in :mod:`examples.phase.simple`, we will examine a 1D problem

.. index::
   single: Grid1D

>>> from fipy import CellVariable, Variable, Grid1D, TransientTerm, DiffusionTerm, ImplicitSourceTerm, PowerLawConvectionTerm, DefaultAsymmetricSolver, Viewer
>>> from fipy.tools import numerix

>>> nx = 400
>>> dx = 5e-6 # cm
>>> L = nx * dx
>>> mesh = Grid1D(dx=dx, nx=nx)

The Helmholtz free energy functional can be written as the integral
:cite:`BoettingerReview:2002` :cite:`McFaddenReview:2002` :cite:`Wheeler:1992`

.. math::

   \\mathcal{F}\\left(\\phi, C, T\\right)
   = \\int_\\mathcal{V} \\left\\{
       f(\\phi, C, T)
       + \\frac{\\kappa_\\phi}{2}\\abs{\\nabla\\phi}^2
       + \\frac{\\kappa_C}{2}\\abs{\\nabla C}^2
   \\right\\} dV

over the volume :math:`\\mathcal{V}` as a function of phase :math:`\\phi` [#phi]_

.. index::
   single: CellVariable

>>> phase = CellVariable(name="phase", mesh=mesh, hasOld=1)

composition :math:`C`

>>> C = CellVariable(name="composition", mesh=mesh, hasOld=1)

and temperature :math:`T` [#T]_

.. index::
   single: Variable

>>> T = Variable(name="temperature")

Frequently, the gradient energy term in concentration is ignored and we
can derive governing equations

.. math::
   :label: eq:phase:binary:phase

    \\frac{\\partial\\phi}{\\partial t}
    = M_\\phi \\left( \\kappa_\\phi \\nabla^2 \\phi
                   - \\frac{\\partial f}{\\partial \\phi} \\right)

for phase and

.. math::
   :label: eq:phase:binary:diffusion

   \\frac{\\partial C}{\\partial t}
   = \\nabla\\cdot\\left( M_C \\nabla \\frac{\\partial f}{\\partial C} \\right)

for solute.

The free energy density :math:`f(\\phi, C, T)` can be constructed in many
different ways. One approach is to construct free energy densities for
each of the pure components, as functions of phase, *e.g.*

.. math::

   f_A(\\phi, T) = p(\\phi) f_A^S(T)
   + \\left(1 - p(\\phi)\\right) f_A^L(T) + \\frac{W_A}{2} g(\\phi)

where :math:`f_A^L(T)`, :math:`f_B^L(T)`, :math:`f_A^S(T)`, and :math:`f_B^S(T)`
are the free energy densities of the pure components. There are a
variety of choices for the interpolation function :math:`p(\\phi)` and the
barrier function :math:`g(\\phi)`,

such as those shown in :mod:`examples.phase.simple`

>>> def p(phi):
...     return phi**3 * (6 * phi**2 - 15 * phi + 10)

>>> def g(phi):
...     return (phi * (1 - phi))**2

The desired thermodynamic model can then be applied to obtain
:math:`f(\\phi, C, T)`, such as for a regular solution,

.. math::

   f(\\phi, C, T) &= (1 - C) f_A(\\phi, T) + C f_B(\\phi, T) \\\\
   &\\qquad + R T \\left[
       (1 - C) \\ln (1 - C) + C \\ln C
   \\right]
   + C (1 - C) \\left[
       \\Omega_S p(\\phi)
       + \\Omega_L \\left( 1 - p(\\phi) \\right)
   \\right]

where

>>> R = 8.314 # J / (mol K)

is the gas constant and :math:`\\Omega_S` and :math:`\\Omega_L` are the
regular solution interaction parameters for solid and liquid.

Another approach is useful when the free energy densities :math:`f^L(C, T)`
and :math:`f^S(C,T)` of the alloy in the solid and liquid phases are
known. This might be the case when the two different phases have
different thermodynamic models or when one or both is obtained from a
Calphad code. In this case, we can construct

.. math::

   f(\\phi, C, T) = p(\\phi) f^S(C,T)
   + \\left(1 - p(\\phi)\\right) f^L(C, T)
   + \\left[
       (1-C) \\frac{W_A}{2} + C \\frac{W_B}{2}
   \\right] g(\\phi).

When the thermodynamic models are the same in both phases, both
approaches should yield the same result.

We choose the first approach and make the simplifying assumptions of an
ideal solution and that

.. math::

   f_A^L(T) & = 0 \\\\
   f_A^S(T) - f_A^L(T) &= \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}

and likewise for component :math:`B`.

>>> LA = 2350. # J / cm**3
>>> LB = 1728. # J / cm**3
>>> TmA = 1728. # K
>>> TmB = 1358. # K

>>> enthalpyA = LA * (T - TmA) / TmA
>>> enthalpyB = LB * (T - TmB) / TmB

This relates the difference between the free energy densities of the
pure solid and pure liquid phases to the latent heat :math:`L_A` and the
pure component melting point :math:`T_M^A`, such that

.. math::

   f_A(\\phi, T) = \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A} p(\\phi)
   + \\frac{W_A}{2} g(\\phi).

With these assumptions

.. math::
   :label: eq:phase:binary:phaseTransformation

   \\frac{\\partial f}{\\partial \\phi}
   &= (1-C) \\frac{\\partial f_A}{\\partial \\phi}
   + C \\frac{\\partial f_B}{\\partial \\phi} \\nonumber \\\\
   &= \\left\\{
       (1-C) \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}
       + C \\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}
   \\right\\} p'(\\phi)
   + \\left\\{
     (1-C) \\frac{W_A}{2} + C \\frac{W_B}{2}
   \\right\\} g'(\\phi)

and

.. math::
   :label: eq:phase:binary:chemicalPotential

   \\frac{\\partial f}{\\partial C}
   &= \\left[f_B(\\phi, T) + \\frac{R T}{V_m} \\ln C\\right]
   - \\left[f_A(\\phi, T) + \\frac{R T}{V_m} \\ln (1-C) \\right] \\nonumber \\\\
   &= \\left[\\mu_B(\\phi, C, T) - \\mu_A(\\phi, C, T) \\right] / V_m

where :math:`\\mu_A` and :math:`\\mu_B` are the classical chemical potentials
for the binary species. :math:`p'(\\phi)` and :math:`g'(\\phi)` are the
partial derivatives of of :math:`p` and :math:`g` with respect to :math:`\\phi`

>>> def pPrime(phi):
...     return 30. * g(phi)

>>> def gPrime(phi):
...     return 2. * phi * (1 - phi) * (1 - 2 * phi)

:math:`V_m` is the molar volume, which we take to be independent of
concentration and phase

>>> Vm = 7.42 # cm**3 / mol

On comparison with :mod:`examples.phase.simple`, we can see that the
present form of the phase field equation is identical to the one found
earlier, with the source now composed of the concentration-weighted average
of the source for either pure component. We let the pure component barriers
equal the previous value

>>> deltaA = deltaB = 1.5 * dx
>>> sigmaA = 3.7e-5 # J / cm**2
>>> sigmaB = 2.9e-5 # J / cm**2
>>> betaA = 0.33 # cm / (K s)
>>> betaB = 0.39 # cm / (K s)
>>> kappaA = 6 * sigmaA * deltaA # J / cm
>>> kappaB = 6 * sigmaB * deltaB # J / cm
>>> WA = 6 * sigmaA / deltaA # J / cm**3
>>> WB = 6 * sigmaB / deltaB # J / cm**3

and define the averages

>>> W = (1 - C) * WA / 2. + C * WB / 2.
>>> enthalpy = (1 - C) * enthalpyA + C * enthalpyB

We can now linearize the source exactly as before

>>> mPhi = -((1 - 2 * phase) * W + 30 * phase * (1 - phase) * enthalpy)
>>> dmPhidPhi = 2 * W - 30 * (1 - 2 * phase) * enthalpy
>>> S1 = dmPhidPhi * phase * (1 - phase) + mPhi * (1 - 2 * phase)
>>> S0 = mPhi * phase * (1 - phase) - S1 * phase

Using the same gradient energy coefficient and phase field mobility

>>> kappa = (1 - C) * kappaA + C * kappaB
>>> Mphi = TmA * betaA / (6 * LA * deltaA)

we define the phase field equation

>>> phaseEq = (TransientTerm(1/Mphi) == DiffusionTerm(coeff=kappa)
...            + S0 + ImplicitSourceTerm(coeff=S1))

----

When coding explicitly, it is typical to simply write a function to
evaluate the chemical potentials :math:`\\mu_A` and :math:`\\mu_B` and then
perform the finite differences necessary to calculate their gradient and
divergence, e.g.,::

    def deltaChemPot(phase, C, T):
        return ((Vm * (enthalpyB * p(phase) + WA * g(phase)) + R * T * log(1 - C)) -
                (Vm * (enthalpyA * p(phase) + WA * g(phase)) + R * T * log(C)))

    for j in range(faces):
        flux[j] = ((Mc[j+.5] + Mc[j-.5]) / 2) \
          * (deltaChemPot(phase[j+.5], C[j+.5], T) \
            - deltaChemPot(phase[j-.5], C[j-.5], T)) / dx

    for j in range(cells):
        diffusion = (flux[j+.5] - flux[j-.5]) / dx

where we neglect the details of the outer boundaries (``j = 0`` and ``j = N``)
or exactly how to translate ``j+.5`` or ``j-.5`` into an array index,
much less the complexities of higher dimensions. FiPy can handle all of
these issues automatically, so we could just write::

    chemPotA = Vm * (enthalpyA * p(phase) + WA * g(phase)) + R * T * log(C)
    chemPotB = Vm * (enthalpyB * p(phase) + WB * g(phase)) + R * T * log(1-C)
    flux = Mc * (chemPotB - chemPotA).faceGrad
    eq = TransientTerm() == flux.divergence

Although the second syntax would essentially work as written, such an
explicit implementation would be very slow. In order to take advantage
of :term:`FiPy`'s implicit solvers, it is necessary to reduce
Eq. :eq:`eq:phase:binary:diffusion` to the canonical form of
Eq. :eq:`num:gen`, hence we must expand
Eq. :eq:`eq:phase:binary:chemicalPotential` as

.. math::

   \\frac{\\partial f}{\\partial C}
   = \\left[
       \\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}
       - \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}
   \\right] p(\\phi)
   + \\frac{R T}{V_m} \\left[\\ln C - \\ln (1-C)\\right]
   + \\frac{W_B - W_A}{2} g(\\phi)

In either bulk phase, :math:`\\nabla p(\\phi) = \\nabla g(\\phi) = 0`, so
we can then reduce Eq. :eq:`eq:phase:binary:diffusion` to

.. math::
   :label: eq:phase:binary:diffusion:bulk

   \\frac{\\partial C}{\\partial t}
   &= \\nabla\\cdot\\left( M_C \\nabla \\left\\{
       \\frac{R T}{V_m} \\left[\\ln C - \\ln (1-C)\\right]
   \\right\\}
   \\right) \\nonumber \\\\
   &= \\nabla\\cdot\\left[
       \\frac{M_C R T}{C (1-C) V_m} \\nabla C
   \\right]

and, by comparison with Fick's second law

.. math::

   \\frac{\\partial C}{\\partial t}
   = \\nabla\\cdot\\left[D \\nabla C\\right],

we can associate the mobility :math:`M_C` with the intrinsic diffusivity :math:`D` by
:math:`M_C \\equiv D C (1-C) V_m / R T` and write Eq. :eq:`eq:phase:binary:diffusion` as

.. math::
   :label: eq:phase:binary:diffusion:canonical

   \\frac{\\partial C}{\\partial t}
   &= \\nabla\\cdot\\left( D \\nabla C \\right) \\nonumber \\\\
   &\\qquad + \\nabla\\cdot\\left(
   \\frac{D C (1 - C) V_m}{R T}
   \\left\\{
       \\left[
           \\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}
           - \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}
       \\right] \\nabla p(\\phi)
       + \\frac{W_B - W_A}{2} \\nabla g(\\phi)
   \\right\\}
   \\right).

The first term is clearly a :class:`~fipy.terms.diffusionTerm.DiffusionTerm`. The second is less
obvious, but by factoring out :math:`C`, we can see that this is a
:class:`~fipy.terms.convectionTerm.ConvectionTerm` with a velocity

.. math::

   \\vec{u}_\\phi =
   \\frac{D (1 - C) V_m}{R T}
   \\left\\{
       \\left[
           \\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}
           - \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}
       \\right] \\nabla p(\\phi)
       + \\frac{W_B - W_A}{2} \\nabla g(\\phi)
   \\right\\}

due to phase transformation, such that

.. math::

   \\frac{\\partial C}{\\partial t}
   = \\nabla\\cdot\\left( D \\nabla C \\right) + \\nabla\\cdot\\left(C \\vec{u}_\\phi\\right)


or

>>> Dl = Variable(value=1e-5) # cm**2 / s
>>> Ds = Variable(value=1e-9) # cm**2 / s
>>> D = (Ds - Dl) * phase.arithmeticFaceValue + Dl

>>> phaseTransformationVelocity = (((enthalpyB - enthalpyA) * p(phase).faceGrad
...                                + 0.5 * (WB - WA) * g(phase).faceGrad)
...                                * D * (1. - C).harmonicFaceValue * Vm / (R * T))

.. index::
   single: PowerLawConvectionTerm

>>> diffusionEq = (TransientTerm()
...                == DiffusionTerm(coeff=D)
...                + PowerLawConvectionTerm(coeff=phaseTransformationVelocity))

----

We initialize the phase field to a step function in the middle of the domain

>>> phase.setValue(1.)
>>> phase.setValue(0., where=mesh.cellCenters[0] > L/2.)

and start with a uniform composition field :math:`C = 1/2`

>>> C.setValue(0.5)

In equilibrium, :math:`\\mu_A(0, C_L, T) = \\mu_A(1, C_S, T)` and
:math:`\\mu_B(0, C_L, T) = \\mu_B(1, C_S, T)` and, for ideal solutions, we can
deduce the liquidus and solidus compositions as

.. math::

   C_L &= \\frac{1 - \\exp\\left(-\\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}\\frac{V_m}{R T}\\right)}
   {\\exp\\left(-\\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}\\frac{V_m}{R T}\\right)
   - \\exp\\left(-\\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}\\frac{V_m}{R T}\\right)} \\\\
   C_S &= \\exp\\left(-\\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}\\frac{V_m}{R T}\\right) C_L

.. index::
   single: exp

>>> Cl = ((1. - numerix.exp(-enthalpyA * Vm / (R * T)))
...       / (numerix.exp(-enthalpyB * Vm / (R * T)) - numerix.exp(-enthalpyA * Vm / (R * T))))
>>> Cs = numerix.exp(-enthalpyB * Vm / (R * T)) * Cl

The phase fraction is predicted by the lever rule

>>> Cavg = C.cellVolumeAverage
>>> fraction = (Cl - Cavg) / (Cl - Cs)

For the special case of ``fraction = Cavg = 0.5``, a little bit of algebra
reveals that the temperature that leaves the phase fraction unchanged is
given by

>>> T.setValue((LA + LB) * TmA * TmB / (LA * TmB + LB * TmA))

In this simple, binary, ideal solution case, we can derive explicit
expressions for the solidus and liquidus compositions. In general, this may
not be possible or practical. In that event, the root-finding facilities in
SciPy can be used.

We'll need a function to return the two conditions for equilibrium

.. math::

   0 = \\mu_A(1, C_S, T) - \\mu_A(0, C_L, T) &=
   \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A} V_m
   + R T \\ln (1 - C_S) - R T \\ln (1 - C_L) \\\\
   0 = \\mu_B(1, C_S, T) - \\mu_B(0, C_L, T) &=
   \\frac{L_B\\left(T - T_M^B\\right)}{T_M^B} V_m
   + R T \\ln C_S - R T \\ln C_L

.. index::
   single: log
   single: array

>>> def equilibrium(C):
...     return [numerix.array(enthalpyA * Vm
...                           + R * T * numerix.log(1 - C[0])
...                           - R * T * numerix.log(1 - C[1])),
...             numerix.array(enthalpyB * Vm
...                           + R * T * numerix.log(C[0])
...                           - R * T * numerix.log(C[1]))]

and we'll have much better luck if we also supply the Jacobian

.. math::

   \\left[\\begin{matrix}
       \\frac{\\partial(\\mu_A^S - \\mu_A^L)}{\\partial C_S}
       & \\frac{\\partial(\\mu_A^S - \\mu_A^L)}{\\partial C_L} \\\\
       \\frac{\\partial(\\mu_B^S - \\mu_B^L)}{\\partial C_S}
       & \\frac{\\partial(\\mu_B^S - \\mu_B^L)}{\\partial C_L}
   \\end{matrix}\\right]
   =
   R T\\left[\\begin{matrix}
       -\\frac{1}{1-C_S} & \\frac{1}{1-C_L} \\\\
       \\frac{1}{C_S} & -\\frac{1}{C_L}
   \\end{matrix}\\right]

>>> def equilibriumJacobian(C):
...     return R * T * numerix.array([[-1. / (1 - C[0]), 1. / (1 - C[1])],
...                                   [ 1. / C[0],      -1. / C[1]]])

.. index::
   single: SciPy

>>> try:
...     from scipy.optimize import fsolve # doctest: +SCIPY
...     CsRoot, ClRoot = fsolve(func=equilibrium, x0=[0.5, 0.5],
...                             fprime=equilibriumJacobian) # doctest: +SCIPY
... except ImportError:
...     ClRoot = CsRoot = 0
...     print("The SciPy library is not available to calculate the solidus and \
... liquidus concentrations")

>>> print(Cl.allclose(ClRoot)) # doctest: +SCIPY
1
>>> print(Cs.allclose(CsRoot)) # doctest: +SCIPY
1

We plot the result against the sharp interface solution

>>> sharp = CellVariable(name="sharp", mesh=mesh)
>>> x = mesh.cellCenters[0]
>>> sharp.setValue(Cs, where=x < L * fraction)
>>> sharp.setValue(Cl, where=x >= L * fraction)

>>> elapsed = Variable(value=0.) # s

.. index::
   pair: module; fipy.viewers

>>> if __name__ == '__main__':
...     try:
...         from phaseViewer import PhaseViewer
...
...         viewer = PhaseViewer(phase=phase, C=C, sharp=sharp,
...                              elapsed=elapsed,
...                              L=L, deltaA=deltaA,
...                              tmin=1e-5, tmax=300 * 3600,
...                              datamin=0., datamax=1.)
...     except ImportError:
...         viewer = Viewer(vars=(phase, C, sharp),
...                         datamin=0., datamax=1.)
...     viewer.plot()

Because the phase field interface will not move, and because we've seen in
earlier examples that the diffusion problem is unconditionally stable, we
need take only one very large timestep to reach equilibrium

>>> dt = 1.e5

Because the phase field equation is coupled to the composition through
``enthalpy`` and ``W`` and the diffusion equation is coupled to the phase
field through ``phaseTransformationVelocity``, it is necessary sweep this
non-linear problem to convergence. We use the "residual" of the equations
(a measure of how well they think they have solved the given set of linear
equations) as a test for how long to sweep. Because of the
:class:`~fipy.terms.convectionTerm.ConvectionTerm`, the solution matrix for
``diffusionEq`` is asymmetric, so we use a
:class:`~fipy.solvers.DefaultAsymmetricSolver` for this equation.

.. .. index:: DefaultAsymmetricSolver, solve, sweep

We now use the ":meth:`~fipy.terms.term.Term.sweep`" method instead of
":meth:`~fipy.terms.term.Term.solve`" because we require the residual.
The initial residual of the diffusion equation is much larger than the norm
of the right-hand-side vector, so we use `"initial"` tolerance scaling.

>>> solver = DefaultAsymmetricSolver(criterion="initial", tolerance=1e-10)

>>> phase.updateOld()
>>> C.updateOld()
>>> phaseRes = 1e+10
>>> diffRes = 1e+10
>>> while phaseRes > 1e-8 or diffRes > 1e-8 or abs(Cavg.value - 0.5) > 5e-7:
...     phaseRes = phaseEq.sweep(var=phase, dt=dt)
...     diffRes = diffusionEq.sweep(var=C, dt=dt, solver=solver)
>>> from fipy import input
>>> if __name__ == '__main__':
...     viewer.plot()
...     input("Stationary phase field. Press <return> to proceed...")

.. image:: /figures/examples/phase/binary/stationary.*
   :width: 90%
   :align: center
   :alt: phase and composition fields in equilibrium, compared with phase diagram concentrations

We verify that the bulk phases have shifted to the predicted solidus and
liquidus compositions

>>> X = mesh.faceCenters[0]
>>> print(Cs.allclose(C.faceValue[X.value==0], atol=2e-4))
True
>>> print(Cl.allclose(C.faceValue[X.value==L], atol=2e-4))
True

and that the phase fraction remains unchanged

>>> print(fraction.allclose(phase.cellVolumeAverage, atol=2e-4))
1

while conserving mass overall

>>> print(Cavg.allclose(0.5, atol=1e-8))
1

----

We now quench by ten degrees

>>> T.setValue(T() - 10.) # K

>>> sharp.setValue(Cs, where=x < L * fraction)
>>> sharp.setValue(Cl, where=x >= L * fraction)

Because this lower temperature will induce the phase interface to move
(solidify), we will need to take much smaller timesteps (the time scales of
diffusion and of phase transformation compete with each other).

The `CFL limit`_ requires that no interface should advect more than one grid
spacing in a timestep. We can get a rough idea for the maximum timestep we can
take by looking at the velocity of convection induced by phase transformation in
Eq. :eq:`eq:phase:binary:diffusion:canonical`. If we assume that the phase changes from 1 to 0 in a single grid spacing,
that the diffusivity is `Dl` at the interface, and that the term due to the difference in
barrier heights is negligible:

.. math::

   \\vec{u}_\\phi &= \\frac{D_\\phi}{C} \\nabla \\phi
   \\\\
   &\\approx
   \\frac{D_l \\frac{1}{2} V_m}{R T}
   \\left[
       \\frac{L_B\\left(T - T_M^B\\right)}{T_M^B}
       - \\frac{L_A\\left(T - T_M^A\\right)}{T_M^A}
   \\right] \\frac{1}{\\Delta x}
   \\\\
   &\\approx
   \\frac{D_l \\frac{1}{2} V_m}{R T}
   \\left(L_B + L_A\\right) \\frac{T_M^A - T_M^B}{T_M^A + T_M^B}
   \\frac{1}{\\Delta x}
   \\\\
   &\\approx 0.28~\\mathrm{cm / s}

To get a :math:`\\text{CFL} = \\vec{u}_\\phi \\Delta t / \\Delta x < 1`, we need a
time step of about :math:`10^{-5}~\\mathrm{s}`.

>>> dt0 = 1.e-5

>>> for i in range(8):
...     phase.updateOld()
...     C.updateOld()
...     phaseRes = 1e+10
...     diffRes = 1e+10
...     while phaseRes > 5e-8 or diffRes > 1e-14 or abs(Cavg.value - 0.5) > 5e-7:
...         phaseRes = phaseEq.sweep(var=phase, dt=dt0)
...         diffRes = diffusionEq.sweep(var=C, dt=dt0, solver=solver)
...     elapsed.value = (i + 1) * dt0
...     if __name__ == '__main__':
...         viewer.plot()

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Moving phase field. Press <return> to proceed...")

.. image:: /figures/examples/phase/binary/binary-0.000080.*
   :width: 90%
   :align: center
   :alt: phase and composition fields in during solidification, compared with final phase diagram concentrations

We see that the composition on either side of the interface approaches the
sharp-interface solidus and liquidus, but it will take a great many more
timesteps to reach equilibrium. If we waited sufficiently long, we
could again verify the final concentrations and phase fraction against the
expected values.

We can estimate the time to equilibration by examining the time for the
diffusion field to become uniform.  In the liquid, this will take
:math:`\\mathcal{O}((10~\\mathrm{\\mu m})^2 / D_l) =
0.1~\\mathrm{s}` and in the solid
:math:`\\mathcal{O}((10~\\mathrm{\\mu m})^2 / D_s) =
1000~\\mathrm{s}`.

Not wanting to take a hundred-million steps, we employ adaptive time
stepping, using the :term:`steppyngstounes` package.  This package takes
care of many of the messy details of stepping, like overshoot, underflow,
and step size adaptation, while keeping the structure of our solve loop
largely intact.

>>> from steppyngstounes import SequenceStepper, PIDStepper # doctest: +STEPPYNGSTOUNES
>>> from itertools import count

Assuming the process is dominated by diffusion, we can take steps that
increase geometrically.  Since we're unsure if diffusion is the only
process controlling dynamics, we take each increasing step with an adaptive
stepper that uses a `PID controller`_ to keep the equation residuals and
mass conservation within acceptable limits.  The total number of solves is
not strongly sensitive to the number of sweeps, but two sweeps seems to be
both sufficient and efficient.

We'll only advance the step if it's successful, so we need to update the
old values before we get started.

>>> phase.updateOld()
>>> C.updateOld()

>>> if __name__ == '__main__':
...     totaltime = 300 * 3600 # 300 h
... else:
...     totaltime = 32e-5 # 320 us

>>> dt = dt0

>>> for checkpoint in SequenceStepper(start=float(elapsed), stop=totaltime,     # doctest: +STEPPYNGSTOUNES
...                                   sizes=(dt0 * 2**(n/2) for n in count(7))):
...     for step in PIDStepper(start=checkpoint.begin,
...                            stop=checkpoint.end,
...                            size=dt):
...         for sweep in range(2):
...             phaseRes = phaseEq.sweep(var=phase, dt=step.size)
...             diffRes = diffusionEq.sweep(var=C, dt=step.size, solver=solver)
...         err = max(phaseRes / 1e-3,
...                   diffRes / 1e-3,
...                   abs(Cavg.value - 0.5) / 5e-7)
...         if step.succeeded(error=err):
...             phase.updateOld()
...             C.updateOld()
...             elapsed.value = step.end
...         else:
...             phase.value = phase.old
...             C.value = C.old
...     # the last step might have been smaller than possible,
...     # if it was near the end of the checkpoint range
...     dt = step.want
...     _ = checkpoint.succeeded()
...     if __name__ == '__main__':
...         viewer.plot()

>>> from fipy import input
>>> if __name__ == '__main__':
...     input("Re-equilbrated phase field. Press <return> to proceed...")

.. image:: /figures/examples/phase/binary/binary-0.000899.*
   :width: 30%
   :alt: phase and composition fields at t=0.000899, compared with final phase diagram concentrations

.. image:: /figures/examples/phase/binary/binary-8.949963.*
   :width: 30%
   :alt: phase and composition fields at t=8.949963, compared with final phase diagram concentrations

.. image:: /figures/examples/phase/binary/binary-1080000.000000.*
   :width: 30%
   :alt: phase and composition fields at t=1080000, compared with final phase diagram concentrations

The interface moves :math:`\\approx 3.4~\\mathrm{\\mu m}` in
:math:`80~\\mathrm{ms}`, driven by diffusion in the liquid
phase (compare the estimate above of :math:`0.1~\\mathrm{s}`).
For the next
:math:`20~\\mathrm{s}`, the interface stalls while the solute step
trapped in the solid phase diffuses outward
(:math:`(3.4~\\mathrm{\\mu m})^2 / D_s =
\\mathcal{O}(100~\\mathrm{s})`).  Once the solute gradient in the
solid reaches the new position of the interface, the solidification front
begins to move, driven by diffusion in the solid.  When the solute in the
solid becomes uniform, the interface stalls again after :math:`\\approx
4000~\\mathrm{s}`, having moved another
:math:`3.2~\\mathrm{\\mu m}` (recall the estimate of
:math:`1000~\\mathrm{s}` for equilibration in the solid).  After this
point, there is essentially no further motion of the interface and barely
perceptible changes in the concentration field.  The fact that the
interface does not reach the predicted phase fraction is due to the fact
that this phase field model accounts for adsorption at the interface,
resulting in the bulk phases not having exactly the concentrations assumed
in the sharp interface treatment.

.. rubric:: Footnotes

.. [#phi] We will find that we need to "sweep" this non-linear problem
   (see *e.g.* the composition-dependent diffusivity example in
   :mod:`examples.diffusion.mesh1D`), so we declare :math:`\\phi` and :math:`C`
   to retain an "old" value.

.. [#T] we are going to want to
   examine different temperatures in this example, so we declare :math:`T`
   as a :class:`~fipy.variables.variable.Variable`

.. _CFL limit: http://en.wikipedia.org/wiki/Courant-Friedrichs-Lewy_condition

.. _PID controller: https://en.wikipedia.org/wiki/PID_controller
"""


__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



"""Domain geometry and topology
"""
__docformat__ = 'restructuredtext'

from fipy.meshes.factoryMeshes import *
from fipy.meshes.periodicGrid1D import *
from fipy.meshes.periodicGrid2D import *
from fipy.meshes.periodicGrid3D import *
from fipy.meshes.skewedGrid2D import *
from fipy.meshes.tri2D import *
from fipy.meshes.gmshMesh import *



__docformat__ = 'restructuredtext'

from .petscPreconditioner import PETScPreconditioner

__all__ = ["ILUPreconditioner"]

class ILUPreconditioner(PETScPreconditioner):
    """Incomplete LU preconditioner for :class:`~fipy.solvers.petsc.petscSolver.PETScSolver`.
    """

    pctype = "ilu"



__docformat__ = 'restructuredtext'


import json
import logging
import warnings

__all__ = [
    "DivergenceWarning",
    "ConvergenceBase",
    "Convergence",
    "AbsoluteToleranceConvergence",
    "HappyBreakdownConvergence",
    "IteratingConvergence",
    "IterationConvergence",
    "LossOfAccuracyConvergence",
    "RHSZeroConvergence",
    "RelativeToleranceConvergence",
    "Divergence",
    "BreakdownDivergence",
    "IllConditionedDivergence",
    "IterationDivergence",
    "NullDivergence",
    "OutOfRangeDivergence",
    "PreconditioningDivergence",
    "IllConditionedPreconditionerDivergence",
    "StagnatedDivergence",
    "ToleranceDivergence"
]

class _ConvergenceMeta(type):
    # We use __init__ rather than __new__ here because we want
    # to modify attributes of the class *after* they have been
    # created
    # See http://jakevdp.github.io/blog/2012/12/01/a-primer-on-python-metaclasses/
    def __init__(cls, name, bases, dct):
        if not hasattr(cls, "code_registry"):
            # this is the base class, create an empty registry
            # one dictionary for all element types
            cls.code_registry = {}

        if not hasattr(cls, "name_registry"):
            # this is the base class, create an empty registry
            # one dictionary for all element types
            cls.name_registry = {}

        super(_ConvergenceMeta, cls).__init__(name, bases, dct)

        if hasattr(cls, "status_code") and hasattr(cls, "suite"):
            cls.code_registry[(cls.suite, cls.status_code)] = cls

        if hasattr(cls, "status_name") and hasattr(cls, "suite"):
            cls.name_registry[(cls.suite, cls.status_name)] = cls

class DivergenceWarning(UserWarning):
    """Warning raised when solver diverges.
    """

    def __init__(self, divergence):
        msg = "msg={status_name}, code={status_code}, residual={residual}".format(**divergence.info)
        super(DivergenceWarning, self).__init__(msg)

class ConvergenceBase(object, metaclass=_ConvergenceMeta):
    """Information about whether and why a solver converged.

    Attributes
    ----------
    solver : ~fipy.solvers.solver.Solver
        The linear solver that was invoked.
    iterations : int
        The number of linear iterations the solver performed.
    criterion : str
        The :ref:`CONVERGENCE` test used by the solver.
    tolerance_scale : float
        The multiplier applied to the tolerance in order for this solver to
        satisfy `criterion`.
    residual : float
        The unscaled norm of the residual achieved by the solver.
    status_code : int or str
        The canonical return value for this type of convergence.
    status_name : str
        The text representation of `status_code`.
    actual_code : int or str
        The status value actually returned by the solver.
    """

    def __init__(self, solver, iterations, residual, criterion, actual_code=None, **kwargs):
        self.solver = solver
        self.iterations = iterations
        self.criterion = criterion
        self.residual = residual
        if actual_code is None:
            self.actual_code = self.status_code
        else:
            self.actual_code = actual_code

        vars(self).update(kwargs)

        self.log()

    @property
    def info(self):
        info = vars(self).copy()
        info["solver"] = str(info["solver"])
        info.update(vars(self.__class__))
        return {k: v for k, v in info.items() if not k.startswith("_")}

    def log(self, level=logging.DEBUG):
        logger = logging.getLogger(self.__class__.__module__
                                   + "." + self.__class__.__name__)

        logger.log(level, json.dumps(self.info))

    def warn(self):
        pass

    def __str__(self):
        return str(self.info)

class Convergence(ConvergenceBase):
    """Information about why a solver converged.
    """

    message = "User requested convergence criteria is satisfied. Iterations: {0}. Relative error: {1}"

class AbsoluteToleranceConvergence(Convergence):
    """Absolute tolerance satisfied

    ::
        residual < atol * scale
    """
    pass

class HappyBreakdownConvergence(Convergence):
    '''"Exact" solution found and more iterations will just make things worse.
    '''
    pass

class IteratingConvergence(Convergence):
    """Solve still in progress.
    """
    pass

class IterationConvergence(Convergence):
    """Requested iterations complete (and no residual calculated).
    """
    pass

class LossOfAccuracyConvergence(Convergence):
    """Numerical loss of precision occurred.
    """
    pass

class RHSZeroConvergence(Convergence):
    r""":math:`\vec{b} = 0`, so exact solution is :math:`\vec{x} = 0`.
    """
    pass

class RelativeToleranceConvergence(Convergence):
    """Relative tolerance satisfied

    ::
        residual < rtol * scale
    """
    pass

class Divergence(ConvergenceBase):
    """Information about why a solver diverged.
    """

    def warn(self):
        warnings.warn(DivergenceWarning(self), stacklevel=5)

class BreakdownDivergence(Divergence):
    """Method broke down.
    """
    pass

class IllConditionedDivergence(Divergence):
    """Matrix was ill-conditioned.
    """
    pass

class IterationDivergence(Divergence):
    """Exceeded maximum iterations.
    """
    pass

class NullDivergence(Divergence):
    """Breakdown when solving the Hessenberg system within GMRES.
    """
    pass

class OutOfRangeDivergence(Divergence):
    """A value became too small, too large, or invalid.
    """
    pass

class PreconditioningDivergence(Divergence):
    """A problem with the preconditioner.
    """
    pass

class IllConditionedPreconditionerDivergence(PreconditioningDivergence):
    """Preconditioner is ill-conditioned.
    """
    pass

class StagnatedDivergence(Divergence):
    """The method stagnated.
    """
    pass

class ToleranceDivergence(Divergence):
    """Residual norm increased too much.
    """
    pass



# -*- coding: utf-8 -*-
#
# FiPy documentation build configuration file, created by
# sphinx-quickstart on Sat Aug 29 21:50:21 2009.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import pathlib
import sys
sys.path.insert(0, pathlib.Path(__file__).parents[2].resolve().as_posix())
sys.path.append(pathlib.Path('sphinxext').resolve().as_posix())
sys.path.append(pathlib.Path('tutorial').resolve().as_posix())

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.intersphinx',
              'sphinx.ext.todo',
              'sphinx.ext.coverage',
              'sphinx.ext.ifconfig',
              'sphinx.ext.autosummary',
              'sphinx.ext.imgconverter',
              'sphinx.ext.napoleon',
              'sphinx_readme',
              'sphinxcontrib.bibtex',
              'matplotlib.sphinxext.plot_directive']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8'

# The root toctree document.
root_doc = 'contents'

bibtex_bibfiles = ['refs.bib']

# General information about the project.
project = u'FiPy'
copyright = u'works of NIST employees are not not subject to copyright protection in the United States'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

import fipy

# The short X.Y version.
version = fipy.__version__
# The full version, including alpha/beta/rc tags.
release = fipy.__version__

# Without this import, get metric craptons of
#   TypeError("'>' not supported between instances of 'size' and 'int'")
import fipy

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
unused_docs = ['RESOURCES',
               'TODOLIST',
               'VERSION']

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_patterns = ['_build',
                    '_templates',
                    'sphinxext',]

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

autoclass_content = "both"

autodoc_default_options = {
    'member-order': 'alphabetical',
    'special-members': True,
    'inherited-members': True,
    'show-inheritance': True,
    'exclude-members': '__annotations__, __dict__, __init__, __module__, __weakref__'
}

autodoc_mock_imports = ['pyamg', 'pyamgx', 'pysvn', 'PyTrilinos.NOX',
                        'PyTrilinos', 'petsc4py', 'mpi4py',
                        'mayavi', 'enthought']

napoleon_numpy_docstring = True
#
# [@MadPhysicist](https://stackoverflow.com/users/2988730/mad-physicist)
# [CC BY-SA 4.0](https://creativecommons.org/licenses/by-sa/4.0/)
# https://stackoverflow.com/a/66182779
# Corrected for https://github.com/sphinx-doc/sphinx/issues/10963
napoleon_preprocess_types = True
napoleon_type_aliases = {
    'array-like': ':term:`array-like <array_like>`',
    'array_like': ':term:`array_like`'
}

autosummary_generate = True

# ~~enable~~ disable :numref:
#
# note: :numref: doesn't work with .. plot::
# https://github.com/matplotlib/matplotlib/issues/9346
#
# There's a workaround, but it's gross
# https://github.com/sfstoolbox/theory/issues/5
#
# Figure numbers skip by 2? Another .. plot:: glitch?
numfig = False

# numpydoc_show_class_members = False

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'nist'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['_themes']

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}
html_sidebars = {'index': ['indexsidebar.html', 'searchbox.html', 'contact.html'],
                 '**': ['localtoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html', 'contact.html']}

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {
    'index': 'index.html',
}

# If false, no module index is generated.
#html_use_modindex = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'FiPydoc'

mathjax3_config = {
  "loader": {"load": ['[tex]/textmacros']},
  "tex": {"packages": {'[+]': ['textmacros']}}
}

# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    'fncychap': r"\usepackage[PetersLenny]{fncychap}",
    'preamble': r"""
        \usepackage{nist}

        \usepackage[amssymb]{SIunits}
        \usepackage{changepage}

        \DeclareMathOperator{\erf}{erf}
        \DeclareMathOperator{\Ei}{ei}
        \providecommand{\abs}[1]{\lvert#1\rvert}


        \authoraddress{Materials Science and Engineering Division \\
        and the Center for Theoretical and Computational Materials Science \\
        Material Measurement Laboratory}

        \definecolor{redish}{rgb}{0.894,0.122,0.122}
        \definecolor{bluish}{rgb}{0.216,0.188,0.533}

        \newcommand{\fipylogo}{\scalebox{10}{\rotatebox{4}{\textcolor{redish}{\( \varphi \)}}\kern-.70em\raisebox{-.15em}{\textcolor{bluish}{\( \pi\)}}}}

        \usepackage{pdflscape}
    """,
    'printindex': r'\footnotesize\raggedright\printindex',
}

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('manual', 'fipy.tex', 'FiPy Manual',
   r'Jonathan E. Guyer \\ Daniel Wheeler \\ James A. Warren', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_toplevel_sectioning = 'part'

# Goofy `.tex.txt` suffix required to keep files from being submitted to
# PDF build process
# https://www.sphinx-doc.org/en/master/latex.html#the-latex-elements-configuration-setting
latex_additional_files = [
    'nist.sty',
    'figures/nistident_flright_vec.pdf',
    'LICENSE.tex.txt',
    'DISCLAIMER.tex.txt'
]

# Documents to append as an appendix to all manuals.
# latex_appendices = ['docs/source/refs.bib_cited']

# If false, no module index is generated.
#latex_use_modindex = True

# --------------------------------------------------------

# refer to Python, NumPy, SciPy, matplotlib
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'mayavi': ('http://docs.enthought.com/mayavi/mayavi', None),
    'steppyngstounes': ('https://pages.nist.gov/steppyngstounes/en/stable/', None),
}

def skip_numpy_not_numerix(app, what, name, obj, skip, options):
    import types
    import numpy

    screen_types = [types.FunctionType,
                     types.BuiltinFunctionType,
                     type(numpy.dtype),
                     type]

    try:
        # NumPy 1.25 changed how dispatching works
        # which makes many numpy functions not Python functions
        # https://github.com/numpy/numpy/pull/23020
        screen_types.append(numpy.core._multiarray_umath._ArrayFunctionDispatcher)
    except:
        pass

    if type(obj) in screen_types:
        module = getattr(obj, "__module__", "")
        if module is None:
            module = ""
        if not (module.startswith("fipy")
                or module.startswith("package")):
            skip = True
    return skip

def skip_mayavi_baseclass(app, what, name, obj, skip, options):
    return True if name in ["trait_property_changed",
                            "trait_items_event"] else None

def autodoc_skip_member(app, what, name, obj, skip, options):
    return (skip_numpy_not_numerix(app, what, name, obj, skip, options)
            or skip_mayavi_baseclass(app, what, name, obj, skip, options))

def setup(app):
    app.connect('autodoc-skip-member', autodoc_skip_member)
    app.add_css_file('widetable.css')
    
# lifted from astropy/astropy@e68ca1a1

# Enable nitpicky mode - which ensures that all references in the docs
# resolve.

# Lots of things aren't defined, though
# https://stackoverflow.com/a/30624034/2019542

nitpicky = True
nitpick_ignore = []

for line in open('nitpick-exceptions'):
    if line.strip() == "" or line.startswith("#"):
        continue
    dtype, target = line.split(None, 1)
    target = target.strip()
    nitpick_ignore.append((dtype, target))


# sphinx-readme options
# See https://sphinx-readme.readthedocs.io/en/latest/configuration/configuring.html
html_context = {
    'display_github': True,
    'github_user': 'usnistgov',
    'github_repo': 'fipy',
}

html_baseurl = "https://pages.nist.gov/fipy/en/latest/"

readme_src_files = [
    "CHANGELOG.rst",
    "INSTALLATION.rst",
    "README.rst",
]
readme_docs_url_type = "code"

readme_raw_directive = False



import os
import sys
import re
from tempfile import mkstemp
from subprocess import Popen, PIPE
from textwrap import dedent

from fipy.tools.parser import parse
from fipy.tools.numerix import sqrt
from fipy.tests import doctestPlus

import examples.phase.anisotropy

def main():
    numberOfElements = parse('--numberOfElements', action='store',
                             type='int', default=10000)
    N = int(sqrt(numberOfElements))

    steps = parse('--numberOfSteps', action='store',
                  type='int', default=20)

    start = parse('--startingStep', action='store',
                  type='int', default=0)

    cpu0 = parse('--cpuBaseLine', action='store',
                  type='float', default=0.)

    args = tuple(sys.argv[1:])

    dir = os.path.dirname(__file__)

    script = doctestPlus._getScript("examples.phase.anisotropy")

    script = script.replace("__main__",
                            "__DONT_RUN_THIS__")

    script = script.replace("nx = ny = 20",
                            "nx = ny = %d" % N)

    script = script.replace("steps = 10",
                            "steps = %d" % steps)

    fd, path = mkstemp(".py")
    os.write(fd, script)
    os.close(fd)

    cputime_RE = re.compile(r"(\d+:)?(\d+):(\d+(\.\d*)?|\d*\.\d+)([eE][-+]?\d+)?")

    def monitor(p):
        def cputimestring2secs(str):
            m = cputime_RE.match(str)

            secs = 60 * int(m.group(2)) + float(m.group(3))
            if m.group(1) is not None:
                secs += 3600 * int(m.group(1))

            return secs

        rsz = vsz = cpu = -1
        while p.poll() is None:
            try:
                r = Popen(("ps", "-p", str(p.pid), "-o" "rsz,vsz,cputime"),
                          stdout=PIPE,
                          env=os.environ).communicate()[0]

                ps = r.splitlines()[1].split()
                rsz = max(rsz, int(ps[0]) * 1024)
                vsz = max(vsz, int(ps[1]) * 1024)
                cpu = max(cpu, cputimestring2secs(ps[2]))
            except:
                break

        return rsz, vsz, cpu

    if start != 0:
        old = '''
              for i in range(steps):
              '''
        new = '''
              mesh_tmp, phase_tmp, dT_tmp = dump.read("%s/anisotropy-%d.dmp.gz")
              phase.setValue(phase_tmp.value)
              dT.setValue(dT_tmp.value)
              for i in range(steps):
              ''' % (dir, start)
        script = script.replace(dedent(old), dedent(new))

    old = r'''\
          #    \cite{WarrenPolycrystal}.
          '''
    new = r'''\
          #    \cite{WarrenPolycrystal}.

          dump.write((mesh, phase, dT), "%s/anisotropy-%%d.dmp.gz" %% (steps + %d))
          ''' % (dir, start)
    script = script.replace(dedent(old), dedent(new))

    f = open(path, "w")
    f.write(script)
    f.close()

    pyth = sys.executable or "python"

    p = Popen((pyth, path) + args, stdout=PIPE)

    rsz, vsz, cpu = monitor(p)

    for l in p.stdout:
        print(l.rstrip())

    print("-" * 79)

    if steps == 0:
        print("           cpu time: %.9f s / step / cell" % cpu)
        print("max resident memory: %.2f B / cell" % float(rsz))
        print(" max virtual memory: %.2f B / cell" % float(vsz))
    else:
        print("           cpu time: %.9f s / step / cell" % ((cpu - cpu0) / steps / N**2))
        print("max resident memory: %.2f B / cell" % (float(rsz) / N**2))
        print(" max virtual memory: %.2f B / cell" % (float(vsz) / N**2))

    os.remove(path)

if __name__ == "__main__":
    main()



"""
1D Viewer that calculates and plots the position of the phase field
interface.
"""

from fipy import Matplotlib1DViewer, numerix
from scipy.optimize import leastsq
from matplotlib import pyplot as plt

class PhaseViewer(Matplotlib1DViewer):
    def __init__(self, phase, C, sharp, elapsed, L, deltaA, title=None,
                 tmin=None, tmax=None, **kwlimits):
        self.phase = phase
        self.x = self.phase.mesh.cellCenters[0]
        self.elapsed = elapsed
        self.L = L
        self.deltaA = deltaA

        fig = plt.figure(constrained_layout=True, figsize=[4, 6])
        gs = fig.add_gridspec(ncols=1, nrows=3)
        self.viewax = fig.add_subplot(gs[1:, 0])
        self.viewax.set_xlabel(r"$x / \mathrm{cm}$")
        self.posax = fig.add_subplot(gs[0, 0], sharex=self.viewax)
        self.posax.set_ylabel(r"$t / \mathrm{s}$")
        self.posax.label_outer()
        self.posax.set_ylim(tmin, tmax)

        self.times = []
        self.positions = []
        self.line, = self.posax.semilogy([0.], [0.],
                                         color="blue")
        self.bug, = self.posax.semilogy([], [],
                                        linestyle="", marker="o", color="red")

        super(PhaseViewer, self).__init__(vars=(phase, C, sharp),
                                          axes=self.viewax,
                                          **kwlimits)

        self.timer = self.viewax.text(0.00025, 0.2, "t = 0")

    @staticmethod
    def tanhResiduals(p, y, x):
        x0, d = p
        return y - 0.5 * (1 - numerix.tanh((x - x0) / (2*d)))

    def _plot(self):
        (x0_fit, d_fit), msg = leastsq(self.tanhResiduals,
                                       [self.L/2., self.deltaA],
                                       args=(self.phase.globalValue,
                                             self.x.globalValue))
        self.positions.append(x0_fit)
        self.times.append(float(self.elapsed))
        self.line.set_data(self.positions, self.times)
        self.bug.set_data([self.positions[-1]], [self.times[-1]])

        self.timer.set_text(r"$t = {:f}\,\mathrm{{s}}$".format(float(self.elapsed)))

        super(PhaseViewer, self)._plot()



r"""

In this example a phase equation is solved in one dimension with a
misorientation present. The phase equation is given by:

.. math::

   \tau_{\phi} \frac{\partial \phi}{\partial t}
   = \alpha^2 \nabla^2 \phi + \phi ( 1 - \phi ) m_1 ( \phi , T)
   - 2 s \phi | \nabla \theta | - \epsilon^2 \phi | \nabla \theta |^2

where

.. math::

   m_1(\phi, T) = \phi - \frac{1}{2} - T \phi ( 1 - \phi )

The initial conditions are:

.. math::

   \phi &= 1 \qquad \forall x  \\
   \theta &= \begin{cases}
   2 \pi / 3 & \text{for $(x - L / 2)^2 + (y - L / 2)^2 > (L / 4)^2$} \\
   -2 \pi / 3 & \text{for $(x - L / 2)^2 + (y - L / 2)^2 \le (L / 4)^2$}
   \end{cases} \\
   T &= 1 \qquad \forall x

and boundary conditions
:math:`\phi = 1` for :math:`x = 0` and :math:`x = L`.

.. Further details of the numerical method for this problem can be found in
   "Extending Phase Field Models of Solidification to Polycrystalline
   Materials", J. A. Warren *et al.*, *Acta Materialia*, **51** (2003)
   6035-6058.

Here the phase equation is solved with an explicit technique.

The solution is allowed to evolve for ``steps = 100`` time steps.

>>> for step in range(steps):
...     phaseEq.solve(phase, dt=timeStepDuration, solver=solver)

The solution is compared with test data. The test data was created
with a FORTRAN code written by Ryo Kobayashi for phase field
modeling. The following code opens the file :file:`modCircle.gz` extracts the
data and compares it with the ``phase`` variable.

>>> import os
>>> testData = numerix.loadtxt(os.path.splitext(__file__)[0] + '.gz')
>>> print(phase.allclose(testData))
1

"""
__docformat__ = 'restructuredtext'


from fipy import input
from fipy import CellVariable, ModularVariable, Grid2D, TransientTerm, ExplicitDiffusionTerm, ImplicitSourceTerm, Viewer
from fipy.tools import numerix

if __name__ == '__main__':
    steps = 100
else:
    steps = 10

timeStepDuration = 0.02
L = 1.5
nx = 100
ny = 100
temperature = 1.
phaseTransientCoeff = 0.1
epsilon = 0.008
s = 0.01
alpha = 0.015

dx = L / nx
dy = L / ny

mesh = Grid2D(dx, dy, nx, ny)

phase = CellVariable(name = 'PhaseField', mesh = mesh, value = 1.)

theta = ModularVariable(name = 'Theta', mesh = mesh, value = 2. * numerix.pi / 3.)
x, y = mesh.cellCenters
theta.setValue(-2. * numerix.pi / 3., where=(x - L / 2.)**2 + (y - L / 2.)**2 < (L / 4.)**2)

mPhiVar = phase - 0.5 + temperature * phase * (1 - phase)
thetaMag = theta.old.grad.mag
implicitSource = mPhiVar * (phase - (mPhiVar < 0))
implicitSource += (2 * s + epsilon**2 * thetaMag) * thetaMag

phaseEq = TransientTerm(phaseTransientCoeff) == \
          ExplicitDiffusionTerm(alpha**2) \
          - ImplicitSourceTerm(implicitSource) \
          + (mPhiVar > 0) * mPhiVar * phase

from fipy import solver_suite
if solver_suite in ["trilinos"]:
    from fipy import ILUPreconditioner
    preconditioner = ILUPreconditioner()
else:
    preconditioner = "default"
from fipy import DefaultAsymmetricSolver
solver = DefaultAsymmetricSolver(precon=preconditioner)

if __name__ == '__main__':

   phaseViewer = Viewer(vars = phase)
   phaseViewer.plot()
   for step in range(steps):
      phaseEq.solve(phase, dt=timeStepDuration, solver=solver)
      phaseViewer.plot()
   input('finished')



__docformat__ = 'restructuredtext'

from .multilevelPreconditioner import MultilevelPreconditioner

__all__ = ["MultilevelNSSAPreconditioner"]

class MultilevelNSSAPreconditioner(MultilevelPreconditioner):
    """Energy-based minimizing smoothed aggregation preconditioner for :class:`~fipy.solvers.trilinos.trilinosSolver.TrilinosSolver`.

    Suitable for highly convective non-symmetric fluid flow problems.
    """

    @property
    def _parameterList(self):
        return {
            "output": 0,
            "max levels": self.levels,
            "prec type": "MGW",
            "increasing or decreasing": "increasing",
            "aggregation: type": "Uncoupled-MIS",
            "energy minimization: enable": True,
            "eigen-analysis: type": "power-method",
            "eigen-analysis: iterations": 20,
            "smoother: sweeps": 4,
            "smoother: damping factor": 0.67,
            "smoother: pre or post": "post",
            "smoother: type": "symmetric Gauss-Seidel",
            "coarse: type": "Amesos-KLU",
            "coarse: max size": 256
        }



from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(docTestModuleNames = (
            'tri2D',
            'mesh1D',
            'mixedelement'
        ), base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



__docformat__ = 'restructuredtext'

from scipy.sparse.linalg import splu

from fipy.solvers.scipy.scipySolver import ScipySolver
from fipy.tools import numerix
from fipy.tools.timer import Timer

__all__ = ["LinearLUSolver"]

class LinearLUSolver(ScipySolver):
    """Interface to :term:`LU`-factorization in :ref:`SciPy`.
    """

    def _adaptLegacyTolerance(self, L, x, b):
        return self._adaptInitialTolerance(L, x, b)

    def _adaptUnscaledTolerance(self, L, x, b):
        return (1., None)

    def _adaptRHSTolerance(self, L, x, b):
        return (self._rhsNorm(L, x, b), None)

    def _adaptMatrixTolerance(self, L, x, b):
        return (self._matrixNorm(L, x, b), None)

    def _adaptInitialTolerance(self, L, x, b):
        return (self._residualNorm(L, x, b), None)

    def _solve_(self, L, x, b):
        """Solve system of equations posed for SciPy

        Parameters
        ----------
        L : ~scipy.sparse.csr_matrix
            Sparse matrix
        x : ndarray
            Solution vector
        b : ndarray
            Right hand side vector

        Returns
        -------
        x : ndarray
            Solution vector
        """
        self._log.debug("BEGIN precondition")

        with Timer() as t:
            diag = L.diagonal()
            maxdiag = max(numerix.absolute(diag))
            L = L * (1 / maxdiag)
            b = b * (1 / maxdiag)

        self._log.debug("END precondition - {} ns".format(t.elapsed))

        tolerance_scale, _ = self._adaptTolerance(L, x, b)

        self._log.debug("BEGIN solve")

        with Timer() as t:
            LU = splu(L.asformat("csc"),
                      diag_pivot_thresh=maxdiag,
                      relax=1,
                      panel_size=10,
                      permc_spec=3)

            for iteration in range(min(self.iterations, 10)):
                residualVector, residual = self._residualVectorAndNorm(L, x, b)

                if residual <= self.tolerance * tolerance_scale:
                    break

                xError = LU.solve(residualVector)
                x[:] = x - xError

        self._log.debug("END solve - {} ns".format(t.elapsed))

        self._setConvergence(suite="scipy",
                             code=0,
                             iterations=iteration+1,
                             residual=residual)

        self.convergence.warn()

        return x



r"""
How to update scripts from version 2.0 to 3.0.

:term:`FiPy` 3.0 introduces several syntax changes from :term:`FiPy` 2.0. We appreciate that
this is very inconvenient for our users, but we hope you'll agree that the new
syntax is easier to read and easier to use. We assure you that this is not
something we do casually; it has been over two and a half years since our last
incompatible change (when :term:`FiPy` 2.0 superceded :term:`FiPy` 1.0).

All examples included with version 3.0 have been updated to use the new syntax,
but any scripts you have written for :term:`FiPy` 2.0 will need to be updated. A
complete listing of the changes needed to take the :term:`FiPy` examples scripts from
version 2.0 to version 3.0 can be found with

    $ git diff version-2_1 version-3_0 examples/

but we summarize the necessary changes here. If these tips are not sufficient to
make your scripts compatible with :term:`FiPy` 3.0, please don't hesitate to ask for
help on the `mailing list`_.


The following items **must** be changed in your scripts

 * We have reconsidered the change in FiPy 2.0 that included all of the
   functions of the :mod:`~fipy.tools.numerix` module in the :mod:`fipy` namespace.
   You now must be more explicit when referring to any of these functions:

   >>> from fipy import *
   >>> y = numerix.exp(x)

   >>> from fipy.tools.numerix import exp
   >>> y = exp(x)

   We generally use the first, but you may see us import specific functions if
   we feel it improves readability. You should feel free to use whichever form
   you find most comfortable.

   .. note:: the old behavior can be obtained, at least for now, by setting
             the :envvar:`FIPY_INCLUDE_NUMERIX_ALL` environment variable.

 * If your equation contains a :class:`~fipy.terms.transientTerm.TransientTerm`,
   then you must specify the timestep by passing a ``dt=`` argument when calling
   :meth:`~fipy.terms.term.Term.solve` or :meth:`~fipy.terms.term.Term.sweep`.

The remaining changes are not *required*, but they make scripts easier to read
and we recommend them. :term:`FiPy` may issue a :exc:`DeprecationWarning` for some cases,
to indicate that we may not maintain the old syntax indefinitely.

 * "getter" and "setter" methods have been replaced with properties, e.g., use

   >>> x, y = mesh.cellCenters

   instead of

   >>> x, y = mesh.getCellCenters()

 * Boundary conditions are better applied with the
   :meth:`~fipy.variables.cellVariable.CellVariable.constrain` method than with
   the old :class:`~fipy.boundaryConditions.fixedValue.FixedValue` and
   :class:`~fipy.boundaryConditions.fixedFlux.FixedFlux` classes. See
   :ref:`BoundaryConditions`.

 * Individual :class:`~fipy.meshes.mesh.Mesh` classes should be imported
   directly from :mod:`fipy.meshes` and not :mod:`fipy.meshes.numMesh`.

 * The :term:`Gmsh` meshes now have simplified names:
   :class:`~fipy.meshes.gmshMesh.Gmsh2D` instead of
   :class:`~fipy.meshes.gmshMesh.GmshImporter2D`,
   :class:`~fipy.meshes.gmshMesh.Gmsh3D` instead of
   :class:`~fipy.meshes.gmshMesh.GmshImporter3D`, and
   :class:`~fipy.meshes.gmshMesh.Gmsh2DIn3DSpace` instead of
   :class:`~fipy.meshes.gmshMesh.GmshImporter2DIn3DSpace`.

.. _mailing list:         http://www.ctcms.nist.gov/fipy/mail.html
"""
__docformat__ = 'restructuredtext'



__docformat__ = 'restructuredtext'

from package.subpackage.base import Base

class Object(Base):
    def __init__(self, arg1, arg2=None, arg3='string'):
        """
        This method, like all those whose names begin and end with
        "``__``" are special.  You won't ever need to call these
        methods directly, but :term:`Python` will invoke them for you under
        certain circumstances, which are described in the
        `Python Reference Manual: Special Method Names`_.

        As an example, the :meth:`__init__` method is invoked when you create an object, as in::

            obj = Object(arg1=something, arg3=somethingElse, ...)

        .. _`Python Reference Manual: Special Method Names`: https://docs.python.org/3/reference/datamodel.html#special-method-names

        Parameters
        ----------
        arg1
            this argument is required. :term:`Python` supports named arguments,
            so you must either list the value for `arg1`  first::

                obj = Object(val1, val2)

            or you can specify the arguments in any order, as long as they are named::

                obj = Object(arg2=val2, arg1=val1)

        arg2
            this argument may be omitted, in which case it will be
            assigned a default value of ``None``.  If you do not use named
            arguments (and we recommend that you do), all required
            arguments must be specified before any optional arguments.
        arg3
            this argument may be omitted, in which case it will be
            assigned a default value of ``'string'``.
        """
        pass

    def method2(self):
        """
        :class:`Object` provides a new definition for the behavior of
        :meth:`method2`, whereas the behavior of
        :meth:`~package.subpackage.base.Base.method1` is defined by
        :class:`~package.subpackage.base.Base`.
        """
        pass



"""

This input file again solves a 1D diffusion problem as in
`./examples/diffusion/steadyState/mesh1D/input.py`. The difference
being that it uses a triangular mesh loaded in using the Gmsh.

The result is again tested in the same way:

    >>> from fipy import CellVariable, GmshGrid3D, DiffusionTerm

    >>> valueLeft = 0.
    >>> valueRight = 1.

    >>> mesh = GmshGrid3D(dx=1, dy=1, dz=1, nx=20, ny=1, nz=1) # doctest: +GMSH

    >>> var = CellVariable(name = "solution variable",
    ...                    mesh = mesh,
    ...                    value = valueLeft) # doctest: +GMSH

    >>> exteriorFaces = mesh.exteriorFaces # doctest: +GMSH
    >>> xFace = mesh.faceCenters[0] # doctest: +GMSH

    >>> var.constrain(valueLeft, exteriorFaces & (xFace ** 2 < 0.000000000000001)) # doctest: +GMSH
    >>> var.constrain(valueRight, exteriorFaces & ((xFace - 20) ** 2 < 0.000000000000001)) # doctest: +GMSH


    >>> DiffusionTerm().solve(var) # doctest: +GMSH
    >>> Lx = 20
    >>> x = mesh.cellCenters[0] # doctest: +GMSH
    >>> analyticalArray = valueLeft + (valueRight - valueLeft) * x / Lx # doctest: +GMSH
    >>> print(var.allclose(analyticalArray, atol = 0.027)) # doctest: +GMSH
    1

"""

__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())




__docformat__ = 'restructuredtext'

from PyTrilinos import AztecOO

from .trilinosPreconditioner import TrilinosPreconditioner

__all__ = ["JacobiPreconditioner"]

class JacobiPreconditioner(TrilinosPreconditioner):
    """Jacobi preconditioner for :class:`~fipy.solvers.trilinos.trilinosSolver.TrilinosSolver`.
    """

    def _applyToSolver(self, solver, matrix):
        solver.SetAztecOption(AztecOO.AZ_precond, AztecOO.AZ_Jacobi)



r"""

Here we solve the level set equation in two dimension for an interior region. The equation is
given by:

.. math::

   \abs{\nabla \phi} &= 1 \\
   \phi &= 0 \;\; \text{at} \begin{cases}
    \text{$x = \left( d, L - d \right)$ for $d \le y \le L - d$} \\
    \text{$y = \left( d, L - d \right)$ for $d \le x \le L - d$}
    \end{cases}

Do the tests:

>>> var.calcDistanceFunction(order=1) #doctest: +LSM

>>> dX = dx / 2.
>>> dY = dy / 2.
>>> mm = dX * dY / numerix.sqrt(dX**2 + dY**2)
>>> def evalCell(phix, phiy, dx, dy):
...     aa = dy**2 + dx**2
...     bb = -2 * ( phix * dy**2 + phiy * dx**2)
...     cc = dy**2 * phix**2 + dx**2 * phiy**2 - dx**2 * dy**2
...     sqr = numerix.sqrt(bb**2 - 4. * aa * cc)
...     return ((-bb - sqr) / 2. / aa,  (-bb + sqr) / 2. / aa)
>>> v1 = evalCell(dY, dX, dx, dy)[1]
>>> v2 = max(-dY*3, -dX*3)
>>> values = numerix.array((  v1,   dY,   dY,  dY,  v1,
...                           dX,  -mm,   -dY,  -mm,   dX,
...                           dX,  -dX,   -v1,  -dX,   dX,
...                           dX,  -mm,   -dY,  -mm,   dX,
...                           v1,   dY,   dY,  dY,  v1  ))
>>> print(var.allclose(values, atol = 1e-10)) #doctest: +LSM
1

"""
__docformat__ = 'restructuredtext'

from fipy import input
from fipy import Grid2D, DistanceVariable, Viewer
from fipy.tools import numerix, serialComm

dx = 1.
dy = 1.
nx = 5
ny = 5

Lx = nx * dx
Ly = ny * dy

mesh = Grid2D(dx = dx, dy = dy, nx = nx, ny = ny, communicator=serialComm)

var = DistanceVariable(
    name = 'level set variable',
    mesh = mesh,
    value = -1.,
    hasOld = 1
    )

x, y = mesh.cellCenters
var.setValue(1, where=((x < dx) | (x > (Lx - dx))
                       | (y < dy) | (y > (Ly - dy))))

if __name__ == '__main__':
    var.calcDistanceFunction(order=1)
    viewer = Viewer(vars=var, datamin=-5., datamax=5.)
    viewer.plot()
    input('finished')





"""This module provides a generic, high-level interface for
creating shared temporary files.  All of the interfaces
provided by this module can be used without fear of race conditions.
"""

from tempfile import NamedTemporaryFile, template, _TemporaryFileWrapper

from fipy.tools import parallelComm

__all__ = ["SharedTemporaryFile"]

# Silence linting errors that come from following conventions of
# tempfile package.
# pylint: disable=invalid-name
# pylint: disable=redefined-builtin
# pylint: disable=too-many-arguments

import io as _io

def SharedTemporaryFile(
    mode="w+b",
    buffering=-1,
    encoding=None,
    newline=None,
    suffix="",
    prefix=template,
    dir=None,
    delete=True,
    communicator=parallelComm,
):
    """Create a temporary file shared by all MPI ranks.

    The file is created as `NamedTemporaryFile` would do it.
    The name of the returned file-like object is accessible as its
    ``name`` attribute.  The file will be automatically deleted when it
    is closed unless the `delete` argument is set to False.

    >>> from fipy.tools import SharedTemporaryFile, parallelComm
    >>> with SharedTemporaryFile(mode='w+', suffix=".tmp") as tmpFile:
    ...     # write on processor 0
    ...     if parallelComm.procID == 0:
    ...         _ = tmpFile.write("shared text")
    ...
    ...     parallelComm.Barrier()
    ...
    ...     # read on all processors
    ...     _ = tmpFile.seek(0)
    ...     txt = tmpFile.read()
    >>> print(txt)
    shared text

    Parameters
    ----------
    prefix, suffix, dir : str
        As for mkstemp
    mode : str
        The mode argument to io.open (default "w+b")
    buffering : int
        The buffer size argument to io.open (default -1)
    encoding : str or None
        The encoding argument to io.open (default None)
    newline : str or None
        The newline argument to io.open (default None)
    delete : bool
        Whether the file is deleted on close (default True)
    communicator : ~fipy.tools.comms.commWrapper.CommWrapper
        MPI communicator describing ranks to share with.  A duck-typed
        object with `procID` and `Nproc` attributes is sufficient.

    Returns
    -------
    file-like object

    See Also
    --------
    tempfile.NamedTemporaryFile, tempfile.mkstemp, io.open

    """

    if communicator.procID == 0:
        file = NamedTemporaryFile(
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            newline=newline,
            suffix=suffix,
            prefix=prefix,
            dir=dir,
            delete=delete,
        )
        fname = file.name
    else:
        fname = None

    fname = communicator.bcast(fname)

    if communicator.procID != 0:
        file = _io.open(
            fname, mode, buffering=buffering, newline=newline, encoding=encoding
        )
        # let procID 0 handle delete
        file = _TemporaryFileWrapper(file, fname, delete=False)

    return file

def _test():
    import fipy.tests.doctestPlus

    return fipy.tests.doctestPlus.testmod()


if __name__ == "__main__":
    _test()



from fipy.solvers.pyamgx import PyAMGXSolver
from fipy.solvers.pyamgx.preconditioners import JacobiPreconditioner

__all__ = ["LinearCGSolver", "LinearPCGSolver"]

class LinearCGSolver(PyAMGXSolver):
    """Interface to the conjugate gradient (:term:`CG`)
    solver in :ref:`PYAMGX`.

    Uses :class:`~fipy.solvers.pyamgx.preconditioners.JacobiPreconditioner` by default.
    """

    CONFIG_DICT = {
        "config_version": 2,
        "determinism_flag": 1,
        "exception_handling" : 1,
        "solver": {
            "convergence": "RELATIVE_INI_CORE",
            "monitor_residual": 1,
            "solver": "PCG",
            "preconditioner": {
               "solver": "NOSOLVER",
            }
        }
    }

    DEFAULT_PRECONDITIONER = JacobiPreconditioner

    def _canSolveAsymmetric(self):
        return False

LinearPCGSolver = LinearCGSolver



"""
Each chapter describes one of the main sub-packages of the :mod:`fipy`
package. The sub-package :mod:`fipy.package` can be found in the directory
:file:`fipy/package/`. In a few cases, there will be packages within
packages, *e.g.* :mod:`fipy.package.subpackage` located in
:file:`fipy/package/subpackage/`. These sub-sub-packages will not be given
their own chapters; rather, their contents will be described in the chapter
for their containing package.
"""

__docformat__ = 'restructuredtext'




from pyamg import smoothed_aggregation_solver

from ...scipy.preconditioners.scipyPreconditioner import ScipyPreconditioner

__all__ = ["SmoothedAggregationPreconditioner"]

class SmoothedAggregationPreconditioner(ScipyPreconditioner):
    """Preconditioner based on `PyAMG smoothed_aggregation_solver`_ for :class:`~fipy.solvers.scipy.scipySolver.ScipySolver`.

    .. _PyAMG smoothed_aggregation_solver: https://pyamg.readthedocs.io/en/latest/generated/pyamg.aggregation.html#pyamg.aggregation.smoothed_aggregation_solver
    """

    def __init__(self):
        pass

    def _applyToMatrix(self, matrix):
        return smoothed_aggregation_solver(matrix).aspreconditioner(cycle='V'), None



"""
An interesting problem is to solve an equation on a 2D geometry that
is embedded in 3D space, such as diffusion on the surface of a sphere
(with nothing either inside or outside the sphere). This example
demonstrates how to create the required mesh.

    >>> from fipy import Gmsh2DIn3DSpace, CellVariable, MayaviClient
    >>> from fipy.tools import numerix

    >>> mesh = Gmsh2DIn3DSpace('''
    ...     radius = 5.0;
    ...     cellSize = 0.3;
    ...
    ...     // create inner 1/8 shell
    ...     Point(1) = {0, 0, 0, cellSize};
    ...     Point(2) = {-radius, 0, 0, cellSize};
    ...     Point(3) = {0, radius, 0, cellSize};
    ...     Point(4) = {0, 0, radius, cellSize};
    ...     Circle(1) = {2, 1, 3};
    ...     Circle(2) = {4, 1, 2};
    ...     Circle(3) = {4, 1, 3};
    ...     Line Loop(1) = {1, -3, 2} ;
    ...     Ruled Surface(1) = {1};
    ...
    ...     // create remaining 7/8 inner shells
    ...     t1[] = Rotate {{0,0,1},{0,0,0},Pi/2} {Duplicata{Surface{1};}};
    ...     t2[] = Rotate {{0,0,1},{0,0,0},Pi} {Duplicata{Surface{1};}};
    ...     t3[] = Rotate {{0,0,1},{0,0,0},Pi*3/2} {Duplicata{Surface{1};}};
    ...     t4[] = Rotate {{0,1,0},{0,0,0},-Pi/2} {Duplicata{Surface{1};}};
    ...     t5[] = Rotate {{0,0,1},{0,0,0},Pi/2} {Duplicata{Surface{t4[0]};}};
    ...     t6[] = Rotate {{0,0,1},{0,0,0},Pi} {Duplicata{Surface{t4[0]};}};
    ...     t7[] = Rotate {{0,0,1},{0,0,0},Pi*3/2} {Duplicata{Surface{t4[0]};}};
    ...
    ...     // create entire inner and outer shell
    ...     Surface Loop(100)={1,t1[0],t2[0],t3[0],t7[0],t4[0],t5[0],t6[0]};
    ... ''').extrude(extrudeFunc=lambda r: 1.1 * r) # doctest: +GMSH

    >>> x, y, z = mesh.cellCenters # doctest: +GMSH

    >>> var = CellVariable(mesh=mesh, value=x * y * z, name="x*y*z") # doctest: +GMSH

    >>> if __name__ == '__main__':
    ...     viewer = MayaviClient(vars=var)
    ...     viewer.plot()

   >>> max(numerix.sqrt(x**2 + y**2 + z**2)) < 5.3 # doctest: +GMSH
   True
   >>> min(numerix.sqrt(x**2 + y**2 + z**2)) > 5.2 # doctest: +GMSH
   True

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')



"""Run all the test cases in examples/diffusion/
"""

from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(testModuleNames = (
                                       'steadyState.test',
                                       'explicit.test',
                                       'nthOrder.test'
                                   ),
                                   docTestModuleNames = (
                                       'mesh1D',
                                       'mesh20x20',
                                       'circle',
                                       'circleQuad',
                                       'electrostatics',
                                       'variable',
                                       'anisotropy',
                                       'mesh20x20Coupled'
                                   ),
                                   base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')




from fipy import parallelComm

def main():
    titles = []
    results = []

    titles.append("mpi4py")

    try:
        from mpi4py import MPI

        m4comm = MPI.COMM_WORLD
        results.append("processor %d of %d" % (m4comm.Get_rank(),
                                               m4comm.Get_size()))

    except Exception as e:
        results.append(str(e))



    titles.append("PyTrilinos")

    try:
        from PyTrilinos import Epetra

        epcomm = Epetra.PyComm()

        results.append("processor %d of %d" % (epcomm.MyPID(),
                                               epcomm.NumProc()))
    except Exception as e:
        results.append(str(e))


    titles.append("petsc4py")

    try:
        from petsc4py import PETSc

        pecomm = PETSc.COMM_WORLD

        results.append("processor %d of %d" % (pecomm.rank,
                                               pecomm.size))
    except Exception as e:
        results.append(str(e))


    titles.append("FiPy")

    try:
        from fipy import Grid1D

        mesh = Grid1D(nx=10)

        results.append("%d cells on processor %d of %d" % (mesh.numberOfCells,
                                                           parallelComm.procID,
                                                           parallelComm.Nproc))
    except Exception as e:
        results.append(str(e))



    lengths = [parallelComm.MaxAll(len(s)) for s in results]
    formats = ["{{{0}:^{1}}}".format(i, l) for i, l in enumerate(lengths)]

    if parallelComm.procID == 0:
        print("    ".join(formats).format(*titles))

    parallelComm.Barrier()

    print(" :: ".join(formats).format(*results))

if __name__ == "__main__":
    main()



r"""

This example represents an expanding circular interface with an initial
coverage of surfactant. The rate of expansion is dependent on the
coverage of surfactant, The governing equations are given by:

.. math::

   \dot{\theta} &= -\frac{\dot{r}}{r} \theta \\
   \dot{r} &= k \theta

The solution for these set of equations is given by:

.. math::

   r &= \sqrt{2 k r_0 \theta_0 t + r_0^2} \\
   \theta &= \frac{r_0 \theta_0}{\sqrt{2 k r_0 \theta_0 t + r_0^2}}

The following tests can be performed. First test for global
conservation of surfactant:

>>> surfactantBefore = numerix.sum(surfactantVariable * mesh.cellVolumes)
>>> totalTime = 0
>>> steps = 5
>>> for step in range(steps):
...     velocity.setValue(surfactantVariable.interfaceVar * k)
...     distanceVariable.extendVariable(velocity)
...     timeStepDuration = cfl * dx / velocity.max()
...     distanceVariable.updateOld()
...     advectionEquation.solve(distanceVariable, dt = timeStepDuration)
...     surfactantEquation.solve(surfactantVariable, dt=1)
...     totalTime += timeStepDuration #doctest: +LSM
>>> surfactantEquation.solve(surfactantVariable, dt=1)
>>> surfactantAfter = numerix.sum(surfactantVariable * mesh.cellVolumes)
>>> print(surfactantBefore.allclose(surfactantAfter))
1

Next test for the correct local value of surfactant:

>>> finalRadius = numerix.sqrt(2 * k * initialRadius * initialSurfactantValue * totalTime + initialRadius**2)
>>> answer = initialSurfactantValue * initialRadius / finalRadius
>>> coverage = surfactantVariable.interfaceVar
>>> error = (coverage / answer - 1)**2 * (coverage > 1e-3)
>>> print(numerix.sqrt(numerix.sum(error) / numerix.sum(error > 0)) < 0.04)
1

Test for the correct position of the interface:

>>> x, y = mesh.cellCenters
>>> radius = numerix.sqrt((x - L / 2)**2 + (y - L / 2)**2)
>>> solution = radius - distanceVariable
>>> error = (solution / finalRadius - 1)**2 * (coverage > 1e-3)
>>> print(numerix.sqrt(numerix.sum(error) / numerix.sum(error > 0)) < 0.02) #doctest: +LSM
1

"""
__docformat__ = 'restructuredtext'

from fipy import input
from fipy import CellVariable, SurfactantVariable, Grid2D, DistanceVariable, TransientTerm, ExplicitUpwindConvectionTerm, AdvectionTerm, Viewer
from fipy.tools import numerix

L = 1.
nx = 50
cfl = 0.1
initialRadius = L / 4.
k = 1
dx = L / nx
steps = 20

from fipy.tools import serialComm
mesh = Grid2D(dx = dx, dy = dx, nx = nx, ny = nx, communicator=serialComm)

x, y = mesh.cellCenters
distanceVariable = DistanceVariable(
    name = 'level set variable',
    mesh = mesh,
    value = numerix.sqrt((x - L / 2.)**2 + (y - L / 2.)**2) - initialRadius,
    hasOld = 1)

initialSurfactantValue =  1.

surfactantVariable = SurfactantVariable(
    value = initialSurfactantValue,
    distanceVar = distanceVariable
    )

velocity = CellVariable(
    name = 'velocity',
    mesh = mesh,
    value = 1.,
    )

advectionEquation = TransientTerm() + AdvectionTerm(velocity)

from fipy.variables.surfactantConvectionVariable import SurfactantConvectionVariable
surfactantEquation = TransientTerm() - \
    ExplicitUpwindConvectionTerm(SurfactantConvectionVariable(distanceVariable))

if __name__ == '__main__':

    distanceViewer = Viewer(vars=distanceVariable,
                            datamin=-initialRadius, datamax=initialRadius)
    surfactantViewer = Viewer(vars=surfactantVariable, datamin=0., datamax=100.)
    velocityViewer = Viewer(vars=velocity, datamin=0., datamax=200.)
    distanceViewer.plot()
    surfactantViewer.plot()
    velocityViewer.plot()

    totalTime = 0

    for step in range(steps):
        print('step', step)
        velocity.setValue(surfactantVariable.interfaceVar * k)
        distanceVariable.extendVariable(velocity)
        timeStepDuration = cfl * dx / velocity.max()
        distanceVariable.updateOld()
        advectionEquation.solve(distanceVariable, dt = timeStepDuration)
        surfactantEquation.solve(surfactantVariable, dt=1)

        totalTime += timeStepDuration

        velocityViewer.plot()
        distanceViewer.plot()
        surfactantViewer.plot()

        finalRadius = numerix.sqrt(2 * k * initialRadius * initialSurfactantValue * totalTime + initialRadius**2)
        answer = initialSurfactantValue * initialRadius / finalRadius
        coverage = surfactantVariable.interfaceVar
        error = (coverage / answer - 1)**2 * (coverage > 1e-3)
        print('error', numerix.sqrt(numerix.sum(error) / numerix.sum(error > 0)))




    input('finished')



r"""

To run this example from the base FiPy directory, type::

    $ python examples/meshing/inputGrid2D.py --numberOfElements=X

This example demonstrates how to build a 1D mesh and obtain basic mesh
information. The command line argument, X, controls the number of
elements on the mesh. Firstly parse the command line argument for
`numberOfElements`, with the default set at 100.

   >>> from fipy.tools.parser import parse
   >>> numberOfElements = parse('--numberOfElements', action = 'store', type = 'int', default = 100)

A `Grid2D` object is invoked in the following way,

   >>> from fipy import Grid2D, CellVariable, Viewer
   >>> from fipy.tools import numerix

   >>> nx = int(numerix.sqrt(numberOfElements))
   >>> ny = nx
   >>> dx = 1.
   >>> dy = 1.
   >>> mesh = Grid2D(nx = nx, ny = nx, dx = dx, dy = dy)

Once the mesh has been built information about the mesh can be
obtained.  For example the mesh volumes can be obtained with the
`getCellVolumes()` method.

   >>> vols = mesh.cellVolumes
   >>> numerix.allclose(dx * dy * numerix.ones(nx * ny), vols)
   1

Obtain the number of cells in the mesh

   >>> N = mesh.numberOfCells
   >>> numerix.allclose(N, numberOfElements)
   1

Obtain all the left exterior faces, this is equal to `ny`.

   >>> faces = mesh.facesLeft
   >>> len(faces) == ny
   1

One can view the mesh with the following code,

   >>> if __name__ == '__main__':
   ...     viewer = Viewer(CellVariable(value = 0, mesh = mesh))
   ...     viewer.plot()

"""
__docformat__ = 'restructuredtext'

from fipy import input

def _run():
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript(__name__))

if __name__ == '__main__':
    _run()
    input("finished")



__docformat__ = 'restructuredtext'

from scipy.sparse.linalg import cgs

from fipy.solvers.scipy.scipyKrylovSolver import ScipyKrylovSolver

__all__ = ["LinearCGSSolver"]

class LinearCGSSolver(ScipyKrylovSolver):
    """Interface to the conjugate gradient squared (:term:`CGS`) solver in
    :ref:`SciPy`.

    No preconditioning by default.
    """

    solveFnc = staticmethod(cgs)




"""
This module can be found in the file :file:`package/subpackage/base.py`.  You make it
available to your script by either::

    import package.subpackage.base

in which case you refer to it by its full name of :mod:`package.subpackage.base`, or::

    from package.subpackage import base

in which case you can refer simply to :mod:`~package.subpackage.base`.
"""
__docformat__ = 'restructuredtext'

class Base(object):
    """
    With very few exceptions, the name of a class will be the capitalized
    form of the module it resides in.  Depending on how you imported the
    module above, you will refer to either :class:`package.subpackage.object.Object` or
    :class:`object.Object`.  Alternatively, you can use::

        from package.subpackage.object import Object

    and then refer simply to :class:`~package.subpackage.object.Object`. For
    many classes, there is a shorthand notation::

        from fipy import Object


    :term:`Python` is an object-oriented language and the FiPy framework
    is composed of objects or classes.  Knowledge of object-oriented
    programming (OOP) is not necessary to use either Python or
    FiPy, but a few concepts are useful.  OOP involves two main
    ideas:

    **encapsulation**
      an object binds data with actions or "methods".  In most cases, you will
      not work with an object's data directly; instead, you will set, retrieve,
      or manipulate the data using the object's methods.

      Methods are functions that are attached to objects and that have
      direct access to the data of those objects.  Rather than
      passing the object data as an argument to a function::

          fn(data, arg1, arg2, ...)

      you instruct an object to invoke an appropriate method::

          object.meth(arg1, arg2, ...)

      If you are unfamiliar with object-oriented practices, there probably seems
      little advantage in this reordering.  You will have to trust us that
      the latter is a much more powerful way to do things.

    **inheritance**
      specialized objects are derived or inherited from more general objects.
      Common behaviors or data are defined in base objects and specific
      behaviors or data are either added or modified in derived objects.
      Objects that declare the existence of certain methods, without actually
      defining what those methods do, are called "abstract".  These objects
      exist to define the behavior of a family of objects, but rely on their
      descendants to actually provide that behavior.

      Unlike many object-oriented languages, :term:`Python` does not prevent the
      creation of abstract objects, but we will include a notice like

      .. attention:: This class is abstract. Always create one of its subclasses.

      for abstract classes which should be used for documentation but
      never actually created in a :term:`FiPy` script.
    """
    def __init__(self):
        pass

    def method1(self):
        """
        This is one thing that you can instruct any object that derives from
        :class:`~package.subpackage.base.Base` to do,
        by calling ``myObjectDerivedFromBase.``:meth:`method1`

        Parameters
        ----------
        self : object
            This special argument refers to the object that is being created.

            .. attention::

               *self* is supplied automatically by the :term:`Python` interpreter to all
               methods.  You don't need to (and should not) specify it yourself.
        """
        pass

    def method2(self):
        """
        This is another thing that you can instruct any object that derives from
        :class:`~package.subpackage.base.Base` to do.
        """
        pass



from fipy.tests.doctestPlus import _LateImportDocTestSuite
import fipy.tests.testProgram

def _suite():
    return _LateImportDocTestSuite(docTestModuleNames = (
            'input',
            'input2D',
        ), base = __name__)

if __name__ == '__main__':
    fipy.tests.testProgram.main(defaultTest='_suite')



# -*- coding: utf-8 -*-

r"""

This example tests diffusion-convection for increasing Péclet numbers.
This test case has been introduced because some solvers were not working
with Péclet numbers over 1.

>>> from fipy import CellVariable, Grid1D, TransientTerm, DiffusionTerm, PowerLawConvectionTerm, Viewer
>>> from fipy.tools import numerix

>>> L = 1.
>>> nx = 1000
>>> dx =  L / nx
>>> mesh = Grid1D(dx=dx, nx=nx)

>>> valueLeft = 0.
>>> valueRight = 1.

>>> var = CellVariable(name = "solution variable", mesh=mesh, value=valueLeft)

>>> var.constrain(valueLeft, mesh.facesLeft)
>>> var.constrain(valueRight, mesh.facesRight)

>>> if __name__ == '__main__':
...     viewer = Viewer(vars = var)

>>> convCoeff = 1.0
>>> peclet = 1e-3
>>> allcloseList = []
>>> from fipy import input
>>> while peclet < 1e4:
...     var[:] = valueLeft
...     diffCoeff = convCoeff * dx / peclet
...     eq = (TransientTerm(1e-4)
...           == DiffusionTerm(coeff=diffCoeff)
...           + PowerLawConvectionTerm(coeff=(convCoeff,)))
...     eq.solve(var=var, dt=1.)
...     x = mesh.cellCenters[0]
...     arg0 = -convCoeff * x / diffCoeff
...     arg0 = numerix.where(arg0 < -200, -200, arg0)
...     arg1 = -convCoeff * L / diffCoeff
...     arg1 = (arg1 >= -200) * (arg1 + 200) - 200
...     CC = 1. - numerix.exp(arg0)
...     DD = 1. - numerix.exp(arg1)
...     analyticalArray = CC / DD
...     allcloseList.append(var.allclose(CC / DD, rtol = 1e-2, atol = 1e-2).value)
...     if __name__ == '__main__':
...         viewer.plot()
...         input("Peclet number: " + str(peclet) + ", press key")
...     peclet *= 10

>>> print(allcloseList)
[True, True, True, True, True, True, True]

"""
__docformat__ = 'restructuredtext'

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())



"""

This example solves the steady-state convection-diffusion equation as
described in :mod:`examples.convection.exponential1D.mesh1D` but
uses the :class:`~fipy.terms.powerLawConvectionTerm.PowerLawConvectionTerm` rather than the
:class:`~fipy.terms.exponentialConvectionTerm.ExponentialConvectionTerm`.

>>> from fipy import CellVariable, Grid1D, DiffusionTerm, PowerLawConvectionTerm, Viewer
>>> from fipy.tools import numerix

>>> L = 10.
>>> nx = 1000
>>> mesh = Grid1D(dx = L / nx, nx = nx)

>>> valueLeft = 0.
>>> valueRight = 1.

>>> var = CellVariable(name = "concentration",
...                    mesh = mesh,
...                    value = valueLeft)

>>> var.constrain(valueLeft, mesh.facesLeft)
>>> var.constrain(valueRight, mesh.facesRight)

>>> diffCoeff = 1.
>>> convCoeff = (10.,)

>>> eq = (DiffusionTerm(coeff=diffCoeff)
...       + PowerLawConvectionTerm(coeff=convCoeff))

>>> eq.solve(var = var)

We test the solution against the analytical result:

>>> axis = 0
>>> x = mesh.cellCenters[axis]
>>> CC = 1. - numerix.exp(-convCoeff[axis] * x / diffCoeff)
>>> DD = 1. - numerix.exp(-convCoeff[axis] * L / diffCoeff)
>>> analyticalArray = CC / DD
>>> print(var.allclose(analyticalArray, rtol = 1e-2, atol = 1e-2))
1

If the problem is run interactively, we can view the result:

>>> if __name__ == '__main__':
...     viewer = Viewer(vars = var)
...     viewer.plot()
"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())

    input('finished')




r"""

This example creates a trench with the following zero level set:

.. math::

   \phi \left( x, y \right) = 0 & \text{when $y = L_y / 5$ and $x \ge L_x / 2$} \\
   \phi \left( x, y \right) = 0 & \text{when $L_y / 5 \le y \le 3 Ly / 5$ and $x = L_x / 2$} \\
   \phi \left( x, y \right) = 0 & \text{when $y = 3 Ly / 5$ and $x \le L_x / 2$}

>>> from fipy import CellVariable, Grid2D, DistanceVariable, TransientTerm, FirstOrderAdvectionTerm, AdvectionTerm, Viewer
>>> from fipy.tools import numerix, serialComm

>>> height = 0.5
>>> Lx = 0.4
>>> Ly = 1.
>>> dx = 0.01
>>> velocity = 1.
>>> cfl = 0.1

>>> nx = int(Lx / dx)
>>> ny = int(Ly / dx)
>>> timeStepDuration = cfl * dx / velocity
>>> steps = 200

>>> mesh = Grid2D(dx = dx, dy = dx, nx = nx, ny = ny, communicator=serialComm)

>>> var = DistanceVariable(name = 'level set variable',
...                        mesh = mesh,
...                        value = -1.,
...                        hasOld = 1
...                        )

>>> x, y = mesh.cellCenters
>>> var.setValue(1, where=(y > 0.6 * Ly) | ((y > 0.2 * Ly) & (x > 0.5 * Lx)))

>>> var.calcDistanceFunction() #doctest: +LSM

>>> advEqn = TransientTerm() + FirstOrderAdvectionTerm(velocity)

The trench is then advected with a unit velocity. The following test can be made
for the initial position of the interface:

>>> r1 =  -numerix.sqrt((x - Lx / 2)**2 + (y - Ly / 5)**2)
>>> r2 =  numerix.sqrt((x - Lx / 2)**2 + (y - 3 * Ly / 5)**2)
>>> d = numerix.zeros((len(x), 3), 'd')
>>> d[:, 0] = numerix.where(x >= Lx / 2, y - Ly / 5, r1)
>>> d[:, 1] = numerix.where(x <= Lx / 2, y - 3 * Ly / 5, r2)
>>> d[:, 2] = numerix.where(numerix.logical_and(Ly / 5 <= y, y <= 3 * Ly / 5), x - Lx / 2, d[:, 0])
>>> argmins = numerix.argmin(numerix.absolute(d), axis = 1)
>>> answer = numerix.take(d.ravel(), numerix.arange(len(argmins))*3 + argmins)
>>> print(var.allclose(answer, atol = 1e-1)) #doctest: +LSM
1

Advect the interface and check the position.

>>> if __name__ == '__main__':
...     viewer = Viewer(vars=var, datamin=-0.1, datamax=0.1)
... 
...     viewer.plot()

>>> for step in range(steps):
...     var.updateOld()
...     advEqn.solve(var, dt = timeStepDuration)
...     if __name__ == '__main__':
...         viewer.plot()

>>> distanceMoved = timeStepDuration * steps * velocity
>>> answer = answer - distanceMoved
>>> answer = numerix.where(answer < 0., 0., answer)
>>> var.setValue(numerix.where(var < 0., 0., var))
>>> print(var.allclose(answer, atol = 1e-1)) #doctest: +LSM
1

"""
__docformat__ = 'restructuredtext'

from fipy import input

if __name__ == '__main__':
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript())
    input('finished')



"""

This input file again solves a 1D diffusion problem as in
`./examples/diffusion/steadyState/mesh1D.py`. The difference in this example is
that the solution method is explicit. The equation used is the
`ExplicitDiffusionEquation`. In this case many steps have to be taken to reach
equilibrium. The `timeStepDuration` parameter specifies the size of each time step
and `steps` is the number of time steps.

    >>> dx = 1.
    >>> dy = 1.
    >>> nx = 10
    >>> ny = 1
    >>> valueLeft = 0.
    >>> valueRight = 1.
    >>> timeStepDuration = 0.02
    >>> steps = 10

A loop is required to execute the necessary time steps:

    >>> for step in range(steps):
    ...     eq.solve(var, solver=solver, dt=timeStepDuration)

The result is again tested in the same way:

    >>> Lx = nx * dx
    >>> x = mesh.cellCenters[0]
    >>> print(var.allclose(answer, rtol = 1e-8))
    1

"""

from fipy import input
from fipy import CellVariable, Tri2D, TransientTerm, ExplicitDiffusionTerm, DefaultSolver, Viewer
from fipy.tools import numerix

dx = 1.
dy = 1.
nx = 10
ny = 1
valueLeft = 0.
valueRight = 1.
timeStepDuration = 0.02

mesh = Tri2D(dx, dy, nx, ny)

var = CellVariable(
    name = "concentration",
    mesh = mesh,
    value = valueLeft)

eq = TransientTerm() == ExplicitDiffusionTerm()

solver = DefaultSolver(tolerance=1e-6, iterations=1000)

var.constrain(valueLeft, mesh.facesLeft)
var.constrain(valueRight, mesh.facesRight)

answer = numerix.array([  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                          0.00000000e+00,  0.00000000e+00,  1.58508452e-07,  6.84325019e-04,
                          7.05111362e-02,  7.81376523e-01,  0.00000000e+00,  0.00000000e+00,
                          0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                          0.00000000e+00,  4.99169535e-05,  1.49682805e-02,  3.82262622e-01,
                          0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                          0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  4.06838361e-06,
                          3.67632029e-03,  1.82227062e-01,  0.00000000e+00,  0.00000000e+00,
                          0.00000000e+00,  0.00000000e+00,  0.00000000e+00,  0.00000000e+00,
                          0.00000000e+00,  4.99169535e-05,  1.49682805e-02,  3.82262622e-01])

if __name__ == '__main__':
    steps = 1000

    for step in range(steps):
        eq.solve(var, solver = solver, dt = timeStepDuration)
    print(var)
    viewer = Viewer(vars = var)
    viewer.plot()
    input('finished')



__docformat__ = 'restructuredtext'

from fipy.solvers.petsc.petscKrylovSolver import PETScKrylovSolver

__all__ = ["LinearGMRESSolver"]

class LinearGMRESSolver(PETScKrylovSolver):

    """Interface to the generalized minimal residual (:term:`GMRES`) solver
    in :ref:`PETSC`.
    """
    
    solver = 'gmres'



r"""Tutorial for writing an electrochemical superfill script.

This input file demonstrates how to
create a new superfill script if the existing suite of scripts do
not meet the required needs. It provides the functionality of
:mod:`examples.levelSet.electroChem.simpleTrenchSystem`.

To run this example from the base FiPy directory type::

    $ python examples/levelSet/electroChem/howToWriteAScript.py --numberOfElements=10000 --numberOfSteps=800

at the command line. The results of the simulation will be displayed
and the word ``finished`` in the terminal at the end of the
simulation. To obtain this example in a plain script file in order to
edit and run type::

    $ fipy_copy_script examples/levelSet/electroChem/howToWriteAScript.py myScript.py

in the base :term:`FiPy` directory. The file :file:`myScript.py` will contain the
script.

The following is an explicit explanation of the input commands
required to set up and run the problem. At the top of the file all the
parameter values are set. Their use will be explained during the
instantiation of various objects and are the same as those explained in
:mod:`examples.levelSet.electroChem.simpleTrenchSystem`.

The following parameters (all in S.I. units)  represent,

 * physical constants,

   >>> faradaysConstant = 9.6e4
   >>> gasConstant = 8.314
   >>> transferCoefficient = 0.5

 * properties associated with the catalyst species,

   >>> rateConstant0 = 1.76
   >>> rateConstant3 = -245e-6
   >>> catalystDiffusion = 1e-9
   >>> siteDensity = 9.8e-6

 * properties of the cupric ions,

   >>> molarVolume = 7.1e-6
   >>> charge = 2
   >>> metalDiffusionCoefficient = 5.6e-10

 * parameters dependent on experimental constraints,

   >>> temperature = 298.
   >>> overpotential = -0.3
   >>> bulkMetalConcentration = 250.
   >>> catalystConcentration = 5e-3
   >>> catalystCoverage = 0.

 * parameters obtained from experiments on flat copper electrodes,

   >>> currentDensity0 = 0.26
   >>> currentDensity1 = 45.

 * general simulation control parameters,

   >>> cflNumber = 0.2
   >>> numberOfCellsInNarrowBand = 10
   >>> cellsBelowTrench = 10
   >>> cellSize = 0.1e-7

 * parameters required for a trench geometry,

   >>> trenchDepth = 0.5e-6
   >>> aspectRatio = 2.
   >>> trenchSpacing = 0.6e-6
   >>> boundaryLayerDepth = 0.3e-6

The hydrodynamic boundary layer depth (``boundaryLayerDepth``) is
intentionally small in this example to keep the mesh at a reasonable
size.

Build the mesh:

.. index::
   pair: module; fipy.tools.parser

>>> from fipy.tools.parser import parse
>>> numberOfElements = parse('--numberOfElements', action='store',
...     type='int', default=-1)
>>> numberOfSteps = parse('--numberOfSteps', action='store',
...     type='int', default=2)

.. index::
   single: sqrt
   single: exp

>>> from fipy import *

>>> if numberOfElements != -1:
...     pos = trenchSpacing * cellsBelowTrench / 4 / numberOfElements
...     sqr = trenchSpacing * (trenchDepth + boundaryLayerDepth) \
...           / (2 * numberOfElements)
...     cellSize = pos + numerix.sqrt(pos**2 + sqr)
... else:
...     cellSize = 0.1e-7

>>> yCells = cellsBelowTrench \
...          + int((trenchDepth + boundaryLayerDepth) / cellSize)
>>> xCells = int(trenchSpacing / 2 / cellSize)

.. index::
   single: Grid2D

>>> from .metalIonDiffusionEquation import buildMetalIonDiffusionEquation
>>> from .adsorbingSurfactantEquation import AdsorbingSurfactantEquation

>>> from fipy import serialComm
>>> mesh = Grid2D(dx=cellSize,
...               dy=cellSize,
...               nx=xCells,
...               ny=yCells,
...               communicator=serialComm)

A ``distanceVariable`` object,
:math:`\phi`, is  required to store  the  position of the interface.

The ``distanceVariable`` calculates its value so that it is a distance
function
(*i.e.* holds the distance at any point in the mesh from the electrolyte/metal
interface at :math:`\phi = 0`) and :math:`|\nabla\phi| = 1`.

First, create the :math:`\phi` variable, which is initially set to -1 everywhere.
Create an initial variable,

.. index::
   single: DistanceVariable

>>> narrowBandWidth = numberOfCellsInNarrowBand * cellSize
>>> distanceVar = DistanceVariable(
...    name='distance variable',
...    mesh= mesh,
...    value=-1.,
...    hasOld=1)

The electrolyte region will be the positive region of the domain while the metal
region will be negative.

>>> bottomHeight = cellsBelowTrench * cellSize
>>> trenchHeight = bottomHeight + trenchDepth
>>> trenchWidth = trenchDepth / aspectRatio
>>> sideWidth = (trenchSpacing - trenchWidth) / 2

>>> x, y = mesh.cellCenters
>>> distanceVar.setValue(1., where=(y > trenchHeight)
...                                 | ((y > bottomHeight)
...                                    & (x < xCells * cellSize - sideWidth)))

>>> distanceVar.calcDistanceFunction(order=2) #doctest: +LSM

The ``distanceVariable`` has now been created to mark the interface. Some other
variables need to be created that govern the concentrations of various species.

Create the catalyst surfactant coverage, :math:`\theta`, variable.
This variable influences the deposition rate.

.. index::
   single: SurfactantVariable

>>> catalystVar = SurfactantVariable(
...     name="catalyst variable",
...     value=catalystCoverage,
...     distanceVar=distanceVar)

Create the bulk catalyst concentration, :math:`c_{\theta}`,
in the electrolyte,

.. index::
   single: CellVariable

>>> bulkCatalystVar = CellVariable(
...     name='bulk catalyst variable',
...     mesh=mesh,
...     value=catalystConcentration)

Create the bulk metal ion concentration,
:math:`c_m`, in the electrolyte.

>>> metalVar = CellVariable(
...     name='metal variable',
...     mesh=mesh,
...     value=bulkMetalConcentration)

The following commands build the ``depositionRateVariable``,
:math:`v`. The ``depositionRateVariable`` is given by the following equation.

.. math::

   v = \frac{i \Omega}{n F}

where :math:`\Omega` is the metal molar volume, :math:`n` is the metal ion
charge and :math:`F` is Faraday's constant. The current density is given
by

.. math::

   i = i_0 \frac{c_m^i}{c_m^{\infty}} \exp{ \left( \frac{- \alpha F}{R T} \eta \right) }

where :math:`c_m^i` is the metal ion concentration in the bulk at the
interface, :math:`c_m^{\infty}` is the far-field bulk concentration of
metal ions, :math:`\alpha` is the transfer coefficient, :math:`R` is the gas
constant, :math:`T` is the temperature and :math:`\eta` is the
overpotential. The exchange current density is an empirical
function of catalyst coverage,

.. math::

   i_0(\theta) = b_0 + b_1 \theta

The commands needed to build this equation are,

>>> expoConstant = -transferCoefficient * faradaysConstant \
...                / (gasConstant * temperature)
>>> tmp = currentDensity1 \
...       * catalystVar.interfaceVar
>>> exchangeCurrentDensity = currentDensity0 + tmp
>>> expo = numerix.exp(expoConstant * overpotential)
>>> currentDensity = expo * exchangeCurrentDensity * metalVar \
...                  / bulkMetalConcentration
>>> depositionRateVariable = currentDensity * molarVolume \
...                          / (charge * faradaysConstant)

Build the extension velocity variable :math:`v_{\text{ext}}`. The extension
velocity uses the
``extensionEquation`` to spread the velocity at the interface to the
rest of the domain.

>>> extensionVelocityVariable = CellVariable(
...     name='extension velocity',
...     mesh=mesh,
...     value=depositionRateVariable)

Using the variables created above the governing equations will be
built.  The governing equation for surfactant conservation is given
by,

.. math::

   \dot{\theta} = J v \theta + k c_{\theta}^i (1 - \theta)

where :math:`\theta` is the coverage of catalyst at the interface,
:math:`J` is the curvature of the interface, :math:`v` is the normal velocity
of the interface, :math:`c_{\theta}^i` is the concentration of
catalyst in the bulk at the interface. The value :math:`k` is given
by an empirical function of overpotential,

.. math::

   k = k_0 + k_3 \eta^3

The above equation is represented by the
:class:`~examples.levelSet.electroChem.adsorbingSurfactantEquation.AdsorbingSurfactantEquation`
in :term:`FiPy`:

>>> surfactantEquation = AdsorbingSurfactantEquation(
...     surfactantVar=catalystVar,
...     distanceVar=distanceVar,
...     bulkVar=bulkCatalystVar,
...     rateConstant=rateConstant0 \
...                    + rateConstant3 * overpotential**3)

The variable :math:`\phi` is advected by the
``advectionEquation`` given by,

.. math::

   \frac{\partial \phi}{\partial t} + v_{\text{ext}}|\nabla \phi| = 0

and is set up with the following commands:

.. index::
   single: AdvectionTerm

>>> advectionEquation = TransientTerm() + AdvectionTerm(extensionVelocityVariable)

The diffusion of metal ions from the far field to the interface is
governed by,

.. math::

   \frac{\partial c_m}{\partial t} = \nabla \cdot D \nabla  c_m

where,

.. math::

   D = \begin{cases}
   D_m & \text{when $\phi > 0$,} \\
   0   & \text{when $\phi \le 0$}
   \end{cases}

The following boundary condition applies at :math:`\phi = 0`,

.. math::

   D \hat{n} \cdot \nabla c = \frac{v}{\Omega}.

The metal ion diffusion equation is set up with the following commands.

>>> metalEquation = buildMetalIonDiffusionEquation(
...     ionVar=metalVar,
...     distanceVar=distanceVar,
...     depositionRate=depositionRateVariable,
...     diffusionCoeff=metalDiffusionCoefficient,
...     metalIonMolarVolume=molarVolume,
... )

>>> metalVar.constrain(bulkMetalConcentration, mesh.facesTop)

The surfactant bulk diffusion equation solves the bulk diffusion of a
species with a source term for the jump from the bulk to an interface.
The governing equation is given by,

.. math::

   \frac{\partial c}{\partial t} = \nabla \cdot D \nabla  c

where,

.. math::

   D = \begin{cases}
   D_{\theta} & \text{when $\phi > 0$} \\
   0          & \text{when $\phi \le 0$}
   \end{cases}

The jump condition at the interface is defined by Langmuir
adsorption. Langmuir adsorption essentially states that the ability
for a species to jump from an electrolyte to an interface is
proportional to the concentration in the electrolyte, available site
density and a jump coefficient. The boundary condition
at :math:`\phi = 0` is given by,

.. math::

   D \hat{n} \cdot \nabla c = -k c (1 - \theta).

The surfactant bulk diffusion equation is set up with the following commands.

>>> from .surfactantBulkDiffusionEquation import buildSurfactantBulkDiffusionEquation
>>> bulkCatalystEquation = buildSurfactantBulkDiffusionEquation(
...     bulkVar=bulkCatalystVar,
...     distanceVar=distanceVar,
...     surfactantVar=catalystVar,
...     diffusionCoeff=catalystDiffusion,
...     rateConstant=rateConstant0 * siteDensity
... )

>>> bulkCatalystVar.constrain(catalystConcentration, mesh.facesTop)

If running interactively, create viewers.

.. index::
   single: MayaviSurfactantViewer

>>> if __name__ == '__main__':
...     try:
...         from .mayaviSurfactantViewer import MayaviSurfactantViewer
...         viewer = MayaviSurfactantViewer(distanceVar,
...                                         catalystVar.interfaceVar,
...                                         zoomFactor=1e6,
...                                         datamax=1.0,
...                                         datamin=0.0,
...                                         smooth=1)
...     except:
...         viewer = MultiViewer(viewers=(
...             Viewer(distanceVar, datamin=-1e-9, datamax=1e-9),
...             Viewer(catalystVar.interfaceVar)))
...         from fipy.models.levelSet.surfactant.matplotlibSurfactantViewer import MatplotlibSurfactantViewer
...         viewer = MatplotlibSurfactantViewer(catalystVar.interfaceVar)
... else:
...     viewer = None

The ``levelSetUpdateFrequency`` defines how often to call the
``distanceEquation`` to reinitialize the ``distanceVariable`` to a
distance function.

>>> levelSetUpdateFrequency = int(0.8 * narrowBandWidth \
...                               / (cellSize * cflNumber * 2))

The following loop runs for ``numberOfSteps`` time steps. The time step
is calculated with the CFL number and the maximum extension velocity.
:math:`v` to
:math:`v_\text{ext}` throughout the whole domain using
:math:`\nabla\phi\cdot\nabla v_\text{ext} = 0`.

>>> for step in range(numberOfSteps):
...
...     if viewer is not None:
...         viewer.plot()
...
...     if step % levelSetUpdateFrequency == 0:
...         distanceVar.calcDistanceFunction(order=2)
...
...     extensionVelocityVariable.setValue(depositionRateVariable())
...
...     distanceVar.updateOld()
...     distanceVar.extendVariable(extensionVelocityVariable, order=2)
...     dt = cflNumber * cellSize / extensionVelocityVariable.max()
...     advectionEquation.solve(distanceVar, dt=dt)
...     surfactantEquation.solve(catalystVar, dt=dt)
...     metalEquation.solve(var=metalVar, dt=dt)
...     bulkCatalystEquation.solve(var=bulkCatalystVar, dt=dt, solver=GeneralSolver()) #doctest: +LSM

The following is a short test case. It uses saved data from a
simulation with 5 time steps. It is not a test for accuracy but a way
to tell if something has changed or been broken.

.. index::
   single: loadtxt

>>> import os

>>> filepath = os.path.join(os.path.split(__file__)[0],
...                         "simpleTrenchSystem.gz")
>>> ##numerix.savetxt(filepath, numerix.array(catalystVar))
>>> print(catalystVar.allclose(numerix.loadtxt(filepath), rtol=1e-4)) #doctest: +LSMLIB
1

>>> from fipy import input
>>> if __name__ == '__main__':
...     input('finished')
"""
__docformat__ = 'restructuredtext'

def _run():
    import fipy.tests.doctestPlus
    exec(fipy.tests.doctestPlus._getScript(__name__))

if __name__ == '__main__':
    _run()



from ..convergence import (Convergence, IterationDivergence,
                           BreakdownDivergence)

class SciPy_Convergence(Convergence):
    """
    """
    status_code = 0
    status_name = "SCIPY_SUCCESS"
    suite = "scipy"

class SciPy_BreakdownDivergence(BreakdownDivergence):
    """
    """
    status_code = -1
    status_name = "SCIPY_ILLEGAL/BREAKDOWN"
    suite = "scipy"

class SciPy_IterationDivergence(IterationDivergence):
    """
    """
    status_code = +1
    status_name = "SCIPY_MAXIT"
    suite = "scipy"



from importlib.util import spec_from_file_location, module_from_spec
from pathlib import Path
import sys
import typer
from typing_extensions import Annotated

__all__ = ["app"]

# from https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
def import_from_path(module_name, file_path):
    spec = spec_from_file_location(module_name, file_path)
    module = module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module

def copy_script(
    copy_from: Annotated[str, typer.Argument(help="path and file name containing script to copy")],
    copy_to: Annotated[str, typer.Argument(help="path and file name to save script to")]
):
    copy_from = Path(copy_from)
    copy_to = Path(copy_to)

    if copy_to.exists():
        raise FileExistsError(copy_to)

    import fipy.tests.doctestPlus

    mod = import_from_path("copy_script_module", copy_from)
    script = fipy.tests.doctestPlus._getScript(name="copy_script_module")
    script = f"""
## This script was derived from
## '{copy_from}'

{script}
"""
    with open(copy_to, "w") as f:
        f.write(script)

    print(f"Script code exported from '{copy_from}' to '{copy_to}'")

app = typer.Typer()
app.command()(copy_script)

if __name__ == "__main__":
    app()



from fipy import input
from fipy import CellVariable, Gmsh2D, DiffusionTerm, Viewer
from fipy.tools import numerix
from matplotlib import cm

def main():
    monitor = None

    geo = """\
    c1_x = 6;
    c1_y = 6;
    c2_x = 4;
    c2_y = 4;
    r = 0.2;

    Point(1) = {0, 0, 0, 1.0};
    Point(2) = {10, 0, 0, 1.0};
    Point(3) = {10, 10, 0, 1.0};
    Point(4) = {1, 10, 0, 1.0};
    Point(5) = {0, 10, 0, 1.0};
    Point(6) = {0, 9, 0, 1.0};
    Point(7) = {c1_x, c1_y, 0, 1.0};
    Point(8) = {c1_x, c1_y + r, 0, 1.0};
    Point(9) = {c1_x + r, c1_y, 0, 1.0};
    Point(10) = {c1_x, c1_y - r, 0, 1.0};
    Point(11) = {c1_x - r, c1_y, 0, 1.0};
    Point(12) = {c2_x, c2_y, 0, 1.0};
    Point(13) = {c2_x, c2_y + r, 0, 1.0};
    Point(14) = {c2_x + r, c2_y, 0, 1.0};
    Point(15) = {c2_x, c2_y - r, 0, 1.0};
    Point(16) = {c2_x - r, c2_y, 0, 1.0};
    Line(1) = {1, 2};
    Line(2) = {2, 3};
    Line(3) = {3, 4};
    Line(4) = {4, 5};
    Line(5) = {5, 6};
    Line(6) = {6, 1};
    Circle(7) = {8, 7, 11};
    Circle(8) = {11, 7, 10};
    Circle(9) = {10, 7, 9};
    Circle(10) = {9, 7, 8};
    Circle(11) = {13, 12, 16};
    Circle(12) = {16, 12, 15};
    Circle(13) = {15, 12, 14};
    Circle(14) = {14, 12, 13};
    Line Loop(1) = {1, 2, 3, 4, 5, 6};
    Line Loop(2) = {7, 8, 9, 10};
    Line Loop(3) = {11, 12, 13, 14};
    Plane Surface(1) = {1, 2, 3};
    Plane Surface(2) = {2};
    Plane Surface(3) = {3};

    Physical Line("Ground") = {4, 5, 6};
    Physical Surface("Field") = {1};
    Physical Surface("Anode") = {2};
    Physical Surface("Cathode") = {3};
    """
    for refinement in range(10):
        mesh = Gmsh2D(geo, background=monitor)

        charge = CellVariable(mesh=mesh, name=r"$\rho$", value=0.)
        charge.setValue(+1, where=mesh.physicalCells["Anode"])
        charge.setValue(-1, where=mesh.physicalCells["Cathode"])

        potential = CellVariable(mesh=mesh, name=r"$\psi$")
        potential.constrain(0., where=mesh.physicalFaces["Ground"])

        eq = DiffusionTerm(coeff=1.) == -charge

        res0 = eq.sweep(var=potential)

        res = eq.justResidualVector(var=potential)

        res1 = numerix.L2norm(res)
        res1a = CellVariable(mesh=mesh, value=abs(res))

        res = CellVariable(mesh=mesh, name="residual", value=abs(res) / mesh.cellVolumes**(1./mesh.dim) / 1e-3)

        # want cells no bigger than 1 and no smaller than 0.001
        maxSize = 1.
        minSize = 0.001
        monitor = CellVariable(mesh=mesh, name="monitor", value= 1. / (res + maxSize) +  minSize)

        viewer = Viewer(vars=potential, xmin=3.5, xmax=4.5, ymin=3.5, ymax=4.5)
    #     viewer = Viewer(vars=(potential, charge))
        viewer.plot()

    #     resviewer = Viewer(vars=res1a, log=True, datamin=1e-6, datamax=1e-2, cmap=cm.gray)
    #     monviewer = Viewer(vars=monitor, log=True, datamin=1e-3, datamax=1)

        input("refinement %d, res0: %g, res: %g:%g, N: %d, min: %g, max: %g, avg: %g. Press <return> to proceed..." \
                  % (refinement, res0, res1, res1a.cellVolumeAverage, mesh.numberOfCells, numerix.sqrt(min(mesh.cellVolumes)), numerix.sqrt(max(mesh.cellVolumes)), numerix.mean(numerix.sqrt(mesh.cellVolumes))))

if __name__ == "__main__":
    main()
